"""Types describing StagYY output data."""

from __future__ import annotations

import typing
from dataclasses import dataclass

import numpy as np

if typing.TYPE_CHECKING:
    from numpy.typing import NDArray


@dataclass(frozen=True)
class Varf:
    """Metadata of scalar field."""

    description: str
    """description of field variable."""
    dim: str
    """dimension used to scale to dimensional values."""


@dataclass(frozen=True)
class Field:
    """Scalar field and associated metadata."""

    values: NDArray[np.float64]
    """values of field."""
    meta: Varf
    """metadata."""


@dataclass(frozen=True)
class Varr:
    """Metadata of radial profiles."""

    description: str
    """description of profile variable."""
    kind: str
    """short description"""
    dim: str
    """dimension used to scale to dimensional values."""


@dataclass(frozen=True)
class Rprof:
    """Radial profile with associated radius and metadata."""

    values: NDArray[np.float64]
    """values of profile."""
    rad: NDArray[np.float64]
    """radial position of profile."""
    meta: Varr
    """metadata."""


@dataclass(frozen=True)
class Vart:
    """Metadata of time series."""

    description: str
    """description of time series."""
    kind: str
    """short description"""
    dim: str
    """dimension used to scale to dimensional values."""


@dataclass(frozen=True)
class Tseries:
    """A time series with associated time and metadata."""

    values: NDArray[np.float64]
    """values of time series."""
    time: NDArray[np.float64]
    """time position of time series."""
    meta: Vart
    """metadata."""



"""Computations of phyvars that are not output by StagYY.

Time series are returned along with the time at which the variables are
evaluated. Radial profiles are returned along with the radial positions at
which the variables are evaluated.
"""

from __future__ import annotations

import typing

import numpy as np
from scipy.integrate import cumulative_trapezoid

from .datatypes import Field, Rprof, Tseries, Varf, Varr, Vart
from .error import NotAvailableError

if typing.TYPE_CHECKING:
    from numpy.typing import NDArray

    from .stagyydata import StagyyData
    from .step import Step


def dtime(sdat: StagyyData) -> Tseries:
    """Time increment.

    Compute dt as a function of time.

    Args:
        sdat: a `StagyyData` instance.

    Returns:
        dt and time arrays.
    """
    time = sdat.tseries.time
    return Tseries(np.diff(time), time[:-1], Vart("Time increment dt", "dt", "s"))


def dt_dt(sdat: StagyyData) -> Tseries:
    """Time derivative of temperature.

    Compute dT/dt as a function of time using an explicit Euler scheme.

    Args:
        sdat: a `StagyyData` instance.

    Returns:
        derivative of temperature and time arrays.
    """
    series = sdat.tseries["Tmean"]
    temp = series.values
    time = series.time
    dtdt = np.diff(temp) / np.diff(time)
    return Tseries(dtdt, time[:-1], Vart("Derivative of temperature", r"dT/dt", "K/s"))


def ebalance(sdat: StagyyData) -> Tseries:
    """Energy balance.

    Compute Nu_t - Nu_b + V*dT/dt as a function of time using an explicit
    Euler scheme. This should be zero if energy is conserved.

    Args:
        sdat: a `StagyyData` instance.

    Returns:
        energy balance and time arrays.
    """
    rbot, rtop = sdat.steps[-1].rprofs.bounds
    if rbot != 0:  # spherical
        coefsurf = (rtop / rbot) ** 2
        volume = rbot * ((rtop / rbot) ** 3 - 1) / 3
    else:
        coefsurf = 1.0
        volume = 1.0
    dtdt = dt_dt(sdat)
    ftop = sdat.tseries["ftop"].values * coefsurf
    fbot = sdat.tseries["fbot"].values
    radio = sdat.tseries["H_int"].values
    ebal: NDArray[np.float64] = ftop[1:] - fbot[1:] + volume * (dtdt.values - radio[1:])  # type: ignore
    return Tseries(ebal, dtdt.time, Vart("Energy balance", r"$\mathrm{Nu}$", "1"))


def mobility(sdat: StagyyData) -> Tseries:
    """Plates mobility.

    Compute the ratio vsurf / vrms.

    Args:
        sdat: a `StagyyData` instance.

    Returns:
        mobility and time arrays.
    """
    time = []
    mob = []
    for step in sdat.steps.filter(rprofs=True):
        time.append(step.timeinfo["time"])
        mob.append(step.rprofs["vrms"].values[-1] / step.timeinfo["Vrms"])
    return Tseries(
        np.array(mob), np.array(time), Vart("Plates mobility", "Mobility", "1")
    )


def delta_r(step: Step) -> Rprof:
    """Cell thicknesses.

    Args:
        step: a `Step` of a `StagyyData` instance.

    Returns:
        the thickness of the cells and radius.
    """
    edges = step.rprofs.walls
    meta = Varr("Cell thickness", "dr", "m")
    return Rprof(np.diff(edges), step.rprofs.centers, meta)


def _scale_prof(
    step: Step, rprof: NDArray[np.float64], rad: NDArray[np.float64] | None = None
) -> NDArray[np.float64]:
    """Scale profile to take sphericity into account."""
    rbot, rtop = step.rprofs.bounds
    if rbot == 0:  # not spherical
        return rprof
    if rad is None:
        rad = step.rprofs.centers
    return rprof * (2 * rad / (rtop + rbot)) ** 2  # type: ignore


def diff_prof(step: Step) -> Rprof:
    """Diffusion flux.

    Args:
        step: a `Step` of a `StagyyData` instance.

    Returns:
        the diffusion and radius.
    """
    rbot, rtop = step.rprofs.bounds
    rad = step.rprofs.centers
    tprof = step.rprofs["Tmean"].values
    diff = (tprof[:-1] - tprof[1:]) / (rad[1:] - rad[:-1])
    # assume tbot = 1
    diff = np.insert(diff, 0, (1 - tprof[0]) / (rad[0] - rbot))
    # assume ttop = 0
    diff = np.append(diff, tprof[-1] / (rtop - rad[-1]))
    meta = Varr("Diffusion", "Heat flux", "W/m2")
    return Rprof(diff, step.rprofs.walls, meta)


def diffs_prof(step: Step) -> Rprof:
    """Scaled diffusion flux.

    This computation takes sphericity into account if necessary.

    Args:
        step: a `Step` of a `StagyyData` instance.

    Returns:
        the diffusion and radius.
    """
    rpf = diff_prof(step)
    meta = Varr("Scaled diffusion", "Heat flux", "W/m2")
    return Rprof(_scale_prof(step, rpf.values, rpf.rad), rpf.rad, meta)


def advts_prof(step: Step) -> Rprof:
    """Scaled advection flux.

    This computation takes sphericity into account if necessary.

    Args:
        step: a `Step` of a `StagyyData` instance.

    Returns:
        the scaled advection and radius.
    """
    return Rprof(
        _scale_prof(step, step.rprofs["advtot"].values),
        step.rprofs.centers,
        Varr("Scaled advection", "Heat flux", "W/m2"),
    )


def advds_prof(step: Step) -> Rprof:
    """Scaled downward advection flux.

    This computation takes sphericity into account if necessary.

    Args:
        step: a `Step` of a `StagyyData` instance.

    Returns:
        the scaled downward advection and radius.
    """
    return Rprof(
        _scale_prof(step, step.rprofs["advdesc"].values),
        step.rprofs.centers,
        Varr("Scaled downward advection", "Heat flux", "W/m2"),
    )


def advas_prof(step: Step) -> Rprof:
    """Scaled upward advection flux.

    This computation takes sphericity into account if necessary.

    Args:
        step: a `Step` of a `StagyyData` instance.

    Returns:
        the scaled upward advection and radius.
    """
    return Rprof(
        _scale_prof(step, step.rprofs["advasc"].values),
        step.rprofs.centers,
        Varr("Scaled upward advection", "Heat flux", "W/m2"),
    )


def energy_prof(step: Step) -> Rprof:
    """Total heat flux.

    This computation takes sphericity into account if necessary.

    Args:
        step: a `Step` of a `StagyyData` instance.

    Returns:
        the energy flux and radius.
    """
    diff_p = diffs_prof(step)
    adv_p = advts_prof(step)
    return Rprof(
        diff_p.values + np.append(adv_p.values, 0),  # type: ignore
        diff_p.rad,
        Varr("Total heat flux", "Heat flux", "W/m2"),
    )


def stream_function(step: Step) -> Field:
    """Stream function (2D).

    Args:
        step: a `Step` of a `StagyyData` instance.

    Returns:
        the stream function field.
    """
    if step.geom.twod_yz:
        x_coord = step.geom.y_walls
        v_x = step.fields["v2"].values[0, :, :, 0]
        v_z = step.fields["v3"].values[0, :, :, 0]
        shape = (1, v_x.shape[0], v_x.shape[1], 1)
    elif step.geom.twod_xz and step.geom.cartesian:
        x_coord = step.geom.x_walls
        v_x = step.fields["v1"].values[:, 0, :, 0]
        v_z = step.fields["v3"].values[:, 0, :, 0]
        shape = (v_x.shape[0], 1, v_x.shape[1], 1)
    else:
        raise NotAvailableError(
            "Stream function only implemented in 2D cartesian and spherical annulus"
        )
    psi = np.zeros_like(v_x)
    if step.geom.spherical:  # YZ annulus
        # positions
        r_nc = step.rprofs.centers  # numerical centers
        r_pc = step.geom.r_centers  # physical centers
        r_nw = step.rprofs.walls[:2]  # numerical walls of first cell
        # vz at center of bottom cells
        vz0 = ((r_nw[1] - r_nc[0]) * v_z[:, 0] + (r_nc[0] - r_nw[0]) * v_z[:, 1]) / (
            r_nw[1] - r_nw[0]
        )
        psi[1:, 0] = -cumulative_trapezoid(r_pc[0] ** 2 * vz0, x=x_coord)
        # vx at center
        vxc = (v_x + np.roll(v_x, -1, axis=0)) / 2
        for i_x in range(len(x_coord)):
            psi[i_x, 1:] = psi[i_x, 0] + cumulative_trapezoid(r_pc * vxc[i_x], x=r_nc)
    else:  # assume cartesian geometry
        z_nc = step.geom.r_centers
        z_nw = step.rprofs.walls[:2]
        vz0 = ((z_nw[1] - z_nc[0]) * v_z[:, 0] + (z_nc[0] - z_nw[0]) * v_z[:, 1]) / (
            z_nw[1] - z_nw[0]
        )
        psi[1:, 0] = -cumulative_trapezoid(vz0, x=x_coord)
        # vx at center
        vxc = (v_x + np.roll(v_x, -1, axis=0)) / 2
        for i_x in range(len(x_coord)):
            psi[i_x, 1:] = psi[i_x, 0] + cumulative_trapezoid(vxc[i_x], x=z_nc)
    if step.geom.twod_xz:
        psi = -psi
    psi = np.reshape(psi, shape)
    return Field(psi, Varf("Stream function", "m2/s"))



"""Definition of non-processing subcommands."""

from __future__ import annotations

import typing
from dataclasses import fields
from textwrap import indent

from rich import box
from rich.columns import Columns
from rich.console import Console
from rich.table import Table

from . import __version__, phyvars
from ._helpers import baredoc, walk
from .config import CONFIG_LOCAL, Config
from .stagyydata import StagyyData, _sdat_from_conf

if typing.TYPE_CHECKING:
    from collections.abc import Iterable, Mapping
    from typing import Callable

    from loam.base import Section

    from .datatypes import Field, Rprof, Tseries, Varf, Varr, Vart
    from .step import Step


def info_cmd(conf: Config) -> None:
    """Print basic information about StagYY run."""
    sdat = _sdat_from_conf(conf.core)
    lsnap = sdat.snaps[-1]
    lstep = sdat.steps[-1]
    print(f"StagYY run in {sdat.path}")
    if lsnap.geom.threed:
        dimension = "{0.nxtot} x {0.nytot} x {0.nztot}".format(lsnap.geom)
    elif lsnap.geom.twod_xz:
        dimension = "{0.nxtot} x {0.nztot}".format(lsnap.geom)
    else:
        dimension = "{0.nytot} x {0.nztot}".format(lsnap.geom)
    if lsnap.geom.cartesian:
        print("Cartesian", dimension)
    elif lsnap.geom.cylindrical:
        print("Cylindrical", dimension)
    else:
        print("Spherical", dimension)
    print()
    for step in walk(sdat, conf):
        print(f"Step {step.istep}/{lstep.istep}", end="")
        if step.isnap is not None:
            print(f", snapshot {step.isnap}/{lsnap.isnap}")
        else:
            print()
        series = step.timeinfo.loc[list(conf.info.output)]
        print(indent(series.to_string(header=False), "  "))
        print()


def _layout(
    dict_vars: Mapping[str, Varf | Varr | Vart],
    dict_vars_extra: Mapping[
        str,
        Callable[[Step], Field | Rprof] | Callable[[StagyyData], Tseries],
    ],
) -> Columns:
    """Print nicely [(var, description)] from phyvars."""
    desc = [(v, m.description) for v, m in dict_vars.items()]
    desc.extend((v, baredoc(m)) for v, m in dict_vars_extra.items())
    return Columns(
        renderables=(f"{var}: [dim]{descr}[/dim]" for var, descr in desc),
        padding=(0, 2),
        column_first=True,
    )


def var_cmd(conf: Config) -> None:
    """Print a list of available variables.

    See [stagpy.phyvars][] where the lists of variables organized by command
    are defined.
    """
    console = Console()
    print_all = not any(getattr(conf.var, fld.name) for fld in fields(conf.var))
    if print_all or conf.var.field:
        console.rule("fields", style="magenta")
        console.print(_layout(phyvars.FIELD, phyvars.FIELD_EXTRA))
        console.print()
    if print_all or conf.var.sfield:
        console.rule("surface fields", style="magenta")
        console.print(_layout(phyvars.SFIELD, {}))
        console.print()
    if print_all or conf.var.rprof:
        console.rule("radial profiles", style="magenta")
        console.print(_layout(phyvars.RPROF, phyvars.RPROF_EXTRA))
        console.print()
    if print_all or conf.var.time:
        console.rule("time series", style="magenta")
        console.print(_layout(phyvars.TIME, phyvars.TIME_EXTRA))
        console.print()
    if print_all or conf.var.refstate:
        console.rule("refstate", style="magenta")
        console.print(_layout(phyvars.REFSTATE, {}))
        console.print()


def version_cmd(conf: Config) -> None:
    """Print StagPy version.

    Use `stagpy.__version__` to obtain the version in a script.
    """
    print(f"stagpy version: {__version__}")


def config_pp(subs: Iterable[str], conf: Config) -> None:
    """Pretty print of configuration options.

    Args:
        subs: conf sections to print.
        conf: configuration.
    """
    console = Console()
    for sub in subs:
        table = Table(title=sub, box=box.SIMPLE)
        table.add_column(header="option", no_wrap=True)
        table.add_column(header="doc")
        table.add_column(header="cli", no_wrap=True)
        table.add_column(header="file", no_wrap=True)

        section: Section = getattr(conf, sub)
        for fld in fields(section):
            opt = fld.name
            entry = section.meta_(opt).entry
            table.add_row(
                opt,
                entry.doc,
                "yes" if entry.in_cli else "no",
                "yes" if entry.in_file else "no",
            )

        if table.rows:
            console.print(table)
            console.print()


def config_cmd(conf: Config) -> None:
    """Configuration handling."""
    if conf.config.create:
        conf.default_().to_file_(CONFIG_LOCAL)
    else:
        config_pp((sec.name for sec in fields(conf)), conf)



"""StagYY par file handling."""

from __future__ import annotations

import typing
from dataclasses import dataclass
from pathlib import Path

import f90nml

from .error import NoParFileError

if typing.TYPE_CHECKING:
    from typing import TypeVar

    from f90nml.namelist import Namelist

    T = TypeVar("T")


@dataclass(frozen=True)
class StagyyPar:
    """A Fortran namelist file written for StagYY."""

    nml: Namelist
    root: Path

    @staticmethod
    def _from_file(parfile: Path) -> StagyyPar:
        if not parfile.is_file():
            raise NoParFileError(parfile)
        par = f90nml.read(str(parfile))
        for content in par.values():
            for option, value in content.items():
                try:
                    content[option] = value.strip()
                except AttributeError:
                    pass
        return StagyyPar(nml=par, root=parfile.parent)

    def _update(self, par_new: StagyyPar) -> None:
        for section, content in par_new.nml.items():
            if section in self.nml:
                self.nml[section].update(content)
            else:
                self.nml[section] = content

    @staticmethod
    def from_main_par(parfile: Path, read_parameters_dat: bool = True) -> StagyyPar:
        """Read StagYY namelist `parfile`.

        The namelist is populated in order with:

        - `par_name_defaultparameters` if it is defined in `parfile`;
        - `par_file` itself;
        - `parameters.dat` if it can be found in the StagYY output directories
          and `read_parameters_dat` is `True`.
        """
        par_main = StagyyPar._from_file(parfile)
        if "default_parameters_parfile" in par_main.nml:
            dfltfile = par_main.get(
                "default_parameters_parfile",
                "par_name_defaultparameters",
                "par_defaults",
            )
            par_dflt = StagyyPar._from_file(par_main.root / dfltfile)
            par_dflt._update(par_main)
            par_main = StagyyPar(nml=par_dflt.nml, root=par_main.root)

        if read_parameters_dat:
            outfile = par_main.legacy_output("_parameters.dat")
            if outfile.is_file():
                par_main._update(StagyyPar._from_file(outfile))
            outfile = par_main.h5_output("parameters.dat")
            if outfile.is_file():
                par_main._update(StagyyPar._from_file(outfile))
        return par_main

    def get(self, section: str, option: str, default: T) -> T:
        sec = self.nml.get(section, {})
        return sec.get(option, default)

    def legacy_output(self, suffix: str) -> Path:
        stem = self.get("ioin", "output_file_stem", "output")
        return self.root / (stem + suffix)

    def h5_output(self, filename: str) -> Path:
        h5folder = self.get("ioin", "hdf5_output_folder", "+hdf5")
        return self.root / h5folder / filename



"""Plate analysis."""

from __future__ import annotations

import typing
from contextlib import suppress
from functools import lru_cache

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import colors
from scipy.signal import argrelmax, argrelmin

from . import _helpers, error, field
from ._helpers import saveplot
from .config import Config
from .datatypes import Field, Varf
from .stagyydata import _sdat_from_conf

if typing.TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import TextIO

    from matplotlib.axes import Axes
    from numpy.typing import NDArray

    from .step import Geometry, Step


def _vzcheck(iphis: NDArray[np.intp], snap: Step, vz_thres: float) -> NDArray[np.intp]:
    """Remove positions where vz is below threshold."""
    # verifying vertical velocity
    vzabs = np.abs(snap.fields["v3"].values[0, ..., 0])
    argdel = []
    for i, iphi in enumerate(iphis):
        vzm = np.mean(vzabs[[iphi - 1, iphi, iphi + 1], :])
        if vzm < vz_thres:
            argdel.append(i)
    return np.delete(iphis, argdel)


@lru_cache
def detect_plates(
    snap: Step, vz_thres_ratio: float = 0
) -> tuple[NDArray[np.intp], NDArray[np.intp]]:
    """Detect plate limits using derivative of horizontal velocity.

    This function is cached for convenience.

    Args:
        snap: a `Step` of a `StagyyData` instance.
        vz_thres_ratio: if above zero, an additional check based on the
            vertical velocities is performed.  Limits detected above a region
            where the vertical velocity is below `vz_thres_ratio * mean(vzabs)`
            are ignored.

    Returns:
        itrenches: phi-indices of detected trenches
        iridges: phi-indices of detected ridges
    """
    dvphi = _surf_diag(snap, "dv2").values

    # finding trenches
    dvphi_saturated = np.copy(dvphi)
    max_dvphi = np.amin(dvphi) * 0.2
    dvphi_saturated[dvphi > max_dvphi] = max_dvphi
    trench_span = 15 if snap.sdat.par.get("boundaries", "air_layer", False) else 10
    itrenches = argrelmin(dvphi_saturated, order=trench_span, mode="wrap")[0]

    # finding ridges
    dvphi_saturated = np.copy(dvphi)
    min_dvphi = np.amax(dvphi) * 0.2
    dvphi_saturated[dvphi < min_dvphi] = min_dvphi
    ridge_span = 20
    iridges = argrelmax(dvphi_saturated, order=ridge_span, mode="wrap")[0]

    # elimination of ridges that are too close to a trench
    phi = snap.geom.p_centers
    phi_trenches = phi[itrenches]
    argdel = []
    if itrenches.size and iridges.size:
        for i, iridge in enumerate(iridges):
            mdistance = np.amin(np.abs(phi_trenches - phi[iridge]))
            phi_span = snap.geom.p_walls[-1] - snap.geom.p_walls[0]
            if mdistance < 2.5e-3 * phi_span:
                argdel.append(i)
        if argdel:
            iridges = np.delete(iridges, argdel)

    # additional check on vz
    if vz_thres_ratio > 0:
        r_w = snap.geom.r_walls
        vz_mean = np.sum(snap.rprofs["vzabs"].values * np.diff(r_w)) / (
            r_w[-1] - r_w[0]
        )
        vz_thres = vz_mean * vz_thres_ratio
        itrenches = _vzcheck(itrenches, snap, vz_thres)
        iridges = _vzcheck(iridges, snap, vz_thres)

    return itrenches, iridges


def _plot_plate_limits(
    axis: Axes, trenches: NDArray[np.float64], ridges: NDArray[np.float64]
) -> None:
    """Plot lines designating ridges and trenches."""
    for trench in trenches:
        axis.axvline(x=trench, color="red", ls="dashed", alpha=0.4)
    for ridge in ridges:
        axis.axvline(x=ridge, color="green", ls="dashed", alpha=0.4)


def _annot_pos(
    geom: Geometry, iphi: int
) -> tuple[tuple[float, float], tuple[float, float]]:
    """Position of arrows to mark limit positions."""
    phi = geom.p_centers[iphi]
    rtot = geom.r_walls[-1]
    thick = rtot - geom.rcmb
    r_beg = rtot + 0.02 * thick
    r_end = rtot + 0.35 * thick
    if geom.cartesian:
        return (phi, r_beg), (phi, r_end)
    # spherical to cartesian
    p_beg = r_beg * np.cos(phi), r_beg * np.sin(phi)
    p_end = r_end * np.cos(phi), r_end * np.sin(phi)
    return p_beg, p_end


def _plot_plate_limits_field(axis: Axes, snap: Step, conf: Config) -> None:
    """Plot arrows designating ridges and trenches in 2D field plots."""
    itrenches, iridges = detect_plates(snap, conf.plates.vzratio)
    for itrench in itrenches:
        p_beg, p_end = _annot_pos(snap.geom, itrench)
        axis.annotate(
            "",
            xy=p_beg,
            xytext=p_end,
            arrowprops=dict(facecolor="red", shrink=0.05),
            annotation_clip=False,
        )
    for iridge in iridges:
        p_beg, p_end = _annot_pos(snap.geom, iridge)
        axis.annotate(
            "",
            xy=p_beg,
            xytext=p_end,
            arrowprops=dict(facecolor="green", shrink=0.05),
            annotation_clip=False,
        )


def _isurf(snap: Step) -> int:
    """Return index of surface accounting for air layer."""
    if snap.sdat.par.get("boundaries", "air_layer", False):
        dsa = snap.sdat.par.nml["boundaries"]["air_thickness"]
        # Remove arbitrary margin to be below the surface.
        # Should check if in the thermal boundary layer.
        rtot = snap.geom.r_walls[-1]
        isurf = snap.geom.at_r(rtot - dsa) - 4
    else:
        isurf = -1
    return isurf


def _surf_diag(snap: Step, name: str) -> Field:
    """Get a surface field.

    Can be a sfield, a regular scalar field evaluated at the surface,
    or dv2 (which is dvphi/dphi).
    """
    with suppress(error.UnknownVarError):
        return snap.sfields[name]
    isurf = _isurf(snap)
    with suppress(error.UnknownVarError):
        field = snap.fields[name]
        return Field(field.values[0, :, isurf, 0], field.meta)
    if name == "dv2":
        vphi = snap.fields["v2"].values[0, :, isurf, 0]
        if snap.geom.cartesian:
            dvphi = np.diff(vphi) / np.diff(snap.geom.p_walls)
        else:
            dvphi = np.diff(vphi) / (
                snap.geom.r_centers[isurf] * np.diff(snap.geom.p_walls)
            )
        return Field(dvphi, Varf(r"$dv_\phi/rd\phi$", "1/s"))
    raise error.UnknownVarError(name)


def _continents_location(snap: Step, at_surface: bool = True) -> NDArray[np.bool]:
    """Location of continents as a boolean array.

    If at_surface is True, it is evaluated only at the surface, otherwise it is
    evaluated in the entire domain.
    """
    icont: int | slice
    if at_surface:
        if snap.sdat.par.get("boundaries", "air_layer", False):
            icont = _isurf(snap) - 6
        else:
            icont = -1
    else:
        icont = slice(None)
    csurf = snap.fields["c"].values[0, :, icont, 0]
    air_layer = snap.sdat.par.get("boundaries", "air_layer", False)
    prot_belts = snap.sdat.par.get("continents", "proterozoic_belts", False)
    if air_layer and not prot_belts:
        return (csurf >= 3) & (csurf <= 4)
    elif air_layer and prot_belts:
        return (csurf >= 3) & (csurf <= 5)
    elif snap.sdat.par.get("tracersin", "tracers_weakcrust", False):
        return csurf >= 3
    return csurf >= 2


def plot_at_surface(
    snap: Step,
    names: Sequence[Sequence[Sequence[str]]],
    conf: Config | None,
) -> None:
    """Plot surface diagnostics.

    Args:
        snap: a `Step` of a `StagyyData` instance.
        names: names of requested surface diagnotics. They are organized by
            figures, plots and subplots.  Surface diagnotics can be valid
            surface field names, field names, or `"dv2"` which is d(vphi)/dphi.
        conf: configuration.
    """
    if conf is None:
        conf = Config.default_()
    for vfig in names:
        fig, axes = plt.subplots(
            nrows=len(vfig), sharex=True, figsize=(12, 2 * len(vfig))
        )
        axes = [axes] if len(vfig) == 1 else axes
        fname = "plates_surf_"
        for axis, vplt in zip(axes, vfig):
            fname += "_".join(vplt) + "_"
            label = ""
            for name in vplt:
                field = _surf_diag(snap, name)
                label = field.meta.description
                phi = (
                    snap.geom.p_centers
                    if field.values.size == snap.geom.nptot
                    else snap.geom.p_walls
                )
                axis.plot(phi, field.values, label=label)
                axis.set_ylim([conf.plot.vmin, conf.plot.vmax])
            if conf.plates.continents:
                continents = _continents_location(snap)
                ymin, ymax = axis.get_ylim()
                axis.fill_between(
                    snap.geom.p_centers,
                    ymin,
                    ymax,
                    where=continents,
                    alpha=0.2,
                    facecolor="#8B6914",
                )
                axis.set_ylim([ymin, ymax])
            phi = snap.geom.p_centers
            itrenches, iridges = detect_plates(snap, conf.plates.vzratio)
            _plot_plate_limits(axis, phi[itrenches], phi[iridges])
            if len(vplt) == 1:
                axis.set_ylabel(label)
            else:
                axis.legend()
        axes[-1].set_xlabel(r"$\phi$")
        axes[-1].set_xlim(snap.geom.p_walls[[0, -1]])
        saveplot(conf, fig, fname, snap.isnap)


def _write_trench_diagnostics(
    step: Step, vrms_surf: float, fid: TextIO, conf: Config
) -> None:
    """Print out some trench diagnostics."""
    assert step.isnap is not None
    itrenches, _ = detect_plates(step, conf.plates.vzratio)
    time = step.time * vrms_surf * conf.scaling.ttransit / conf.scaling.yearins / 1.0e6
    isurf = _isurf(step)
    trenches = step.geom.p_centers[itrenches]

    # vphi at trenches
    vphi = step.fields["v2"].values[0, :, isurf, 0]
    vphi: NDArray[np.float64] = (vphi[1:] + vphi[:-1]) / 2  # type: ignore
    v_trenches = vphi[itrenches]

    agetrenches = np.zeros(len(itrenches))
    if "age" in step.fields:
        agefld = step.fields["age"].values[0, :, isurf, 0]
        age_surface = np.ma.masked_where(agefld < 1.0e-5, agefld)
        age_surface_dim = (
            age_surface
            * vrms_surf
            * conf.scaling.ttransit
            / conf.scaling.yearins
            / 1.0e6
        )
        agetrenches = age_surface_dim[itrenches]  # age at the trench

    if conf.plates.continents:
        phi_cont = step.geom.p_centers[_continents_location(step)]
    else:
        phi_cont = np.array([np.nan])

    distance_subd = []
    ph_cont_subd = []
    for trench_i in trenches:
        # compute distance between subduction and continent
        angdistance1 = np.abs(phi_cont - trench_i)
        angdistance2 = 2 * np.pi - angdistance1
        angdistance = np.minimum(angdistance1, angdistance2)
        i_closest = np.argmin(angdistance)

        ph_cont_subd.append(phi_cont[i_closest])
        distance_subd.append(angdistance[i_closest])

    # writing the output into a file, all time steps are in one file
    for isubd in range(len(trenches)):
        fid.write(
            "%6.0f %11.7f %11.3f %10.6f %10.6f %10.6f %10.6f %11.3f\n"
            % (
                step.isnap,
                step.time,
                time,
                trenches[isubd],
                v_trenches[isubd],
                distance_subd[isubd],
                ph_cont_subd[isubd],
                agetrenches[isubd],
            )
        )


def plot_scalar_field(
    snap: Step,
    fieldname: str,
    conf: Config | None = None,
) -> None:
    """Plot scalar field with plate information.

    Args:
        snap: a `Step` of a `StagyyData` instance.
        fieldname: name of the field that should be decorated with plate
            informations.
        conf: configuration.
    """
    if conf is None:
        conf = Config.default_()
    fig, axis, _, _ = field.plot_scalar(snap, fieldname, conf=conf)

    if conf.plates.continents:
        c_field = np.ma.masked_where(
            ~_continents_location(snap, at_surface=False),
            snap.fields["c"].values[0, :, :, 0],
        )
        cmap = colors.ListedColormap(["k", "g", "m"])
        with conf.field.context_(colorbar=False):
            field.plot_scalar(
                snap,
                "c",
                c_field,
                axis,
                conf=conf,
                cmap=cmap,
                norm=colors.BoundaryNorm([2, 3, 4, 5], cmap.N),
            )

    # plotting velocity vectors
    field.plot_vec(axis, snap, "sx" if conf.plates.stress else "v", conf=conf)

    # Put arrow where ridges and trenches are
    _plot_plate_limits_field(axis, snap, conf)

    saveplot(
        conf,
        fig,
        f"plates_{fieldname}",
        snap.isnap,
        close=conf.plates.zoom is None,
    )

    # Zoom
    if conf.plates.zoom is not None:
        if not 0 <= conf.plates.zoom <= 360:
            raise error.InvalidZoomError(conf.plates.zoom)
        if 45 < conf.plates.zoom <= 135:
            ladd, radd, uadd, dadd = 0.8, 0.8, 0.05, 0.1
        elif 135 < conf.plates.zoom <= 225:
            ladd, radd, uadd, dadd = 0.05, 0.1, 0.8, 0.8
        elif 225 < conf.plates.zoom <= 315:
            ladd, radd, uadd, dadd = 0.8, 0.8, 0.1, 0.05
        else:  # >315 or <=45
            ladd, radd, uadd, dadd = 0.1, 0.05, 0.8, 0.8
        xzoom = (snap.geom.rcmb + 1) * np.cos(np.radians(conf.plates.zoom))
        yzoom = (snap.geom.rcmb + 1) * np.sin(np.radians(conf.plates.zoom))
        axis.set_xlim(xzoom - ladd, xzoom + radd)
        axis.set_ylim(yzoom - dadd, yzoom + uadd)
        saveplot(conf, fig, f"plates_zoom_{fieldname}", snap.isnap)


def cmd(conf: Config) -> None:
    """Plate analysis.

    This is the implementation of the `plates` subcommand.
    """
    sdat = _sdat_from_conf(conf.core)
    view = _helpers.walk(sdat, conf)

    isurf = _isurf(next(iter(view)))
    vrms_surf = view.filter(rprofs=True).rprofs_averaged["vhrms"].values[isurf]
    nb_plates = []
    time = []
    istart, iend = None, None

    oname = _helpers.out_name(conf, f"plates_trenches_{view.stepstr}")
    with open(f"{oname}.dat", "w") as fid:
        fid.write(
            "#  istep     time   time_My   phi_trench  vel_trench  "
            + "distance     phi_cont  age_trench_My\n"
        )

        for step in view.filter(fields=["T"]):
            # could check other fields too
            _write_trench_diagnostics(step, vrms_surf, fid, conf)
            plot_at_surface(step, conf.plates.plot, conf)
            plot_scalar_field(step, conf.plates.field, conf)
            if conf.plates.nbplates:
                time.append(step.time)
                itr, ird = detect_plates(step, conf.plates.vzratio)
                nb_plates.append(itr.size + ird.size)
                istart = step.isnap if istart is None else istart
                iend = step.isnap
            if conf.plates.distribution:
                phi = step.geom.p_centers
                itr, ird = detect_plates(step, conf.plates.vzratio)
                limits = np.concatenate((phi[itr], phi[ird]))
                limits.sort()
                plate_sizes = np.diff(limits, append=2 * np.pi + limits[0])
                fig, axis = plt.subplots()
                axis.hist(plate_sizes, bins=10, range=(0, np.pi))
                axis.set_ylabel("Number of plates")
                axis.set_xlabel(r"$\phi$-span")
                saveplot(conf, fig, "plates_size_distribution", step.isnap)

        if conf.plates.nbplates:
            figt, axis = plt.subplots()
            axis.plot(time, nb_plates)
            axis.set_xlabel("Time")
            axis.set_ylabel("Number of plates")
            saveplot(conf, figt, f"plates_{istart}_{iend}")



"""Plot radial profiles."""

from __future__ import annotations

import typing

import matplotlib.pyplot as plt
import numpy as np

from . import _helpers
from .config import Config
from .stagyydata import _sdat_from_conf

if typing.TYPE_CHECKING:
    from collections.abc import Sequence

    from numpy.typing import NDArray

    from .step import Rprofs, Step


def plot_rprofs(
    rprofs: Rprofs,
    names: Sequence[Sequence[Sequence[str]]],
    conf: Config | None = None,
) -> None:
    """Plot requested radial profiles.

    Args:
        rprofs: a radial profile collection, such as `Step.rprofs` or
            [`StepsView.rprofs_averaged`][stagpy.stagyydata.StepsView.rprofs_averaged].
        names: profile names organized by figures, plots and subplots.
        conf: configuration.
    """
    if conf is None:
        conf = Config.default_()
    stepstr = rprofs.stepstr

    for vfig in names:
        fig, axes = plt.subplots(
            ncols=len(vfig), sharey=True, figsize=(4 * len(vfig), 6)
        )
        axes = [axes] if len(vfig) == 1 else axes
        fname = "rprof_"
        for iplt, vplt in enumerate(vfig):
            xlabel = None
            profs_on_plt = (rprofs[rvar] for rvar in vplt)
            fname += "_".join(vplt) + "_"
            for ivar, rpf in enumerate(profs_on_plt):
                rprof = rpf.values
                rad = rpf.rad
                meta = rpf.meta
                if conf.rprof.depth:
                    rad: NDArray[np.float64] = rprofs.bounds[1] - rad  # type: ignore
                axes[iplt].plot(rprof, rad, conf.rprof.style, label=meta.description)
                if xlabel is None:
                    xlabel = meta.kind
                elif xlabel != meta.kind:
                    xlabel = ""
            if ivar == 0:
                xlabel = meta.description
            if xlabel:
                axes[iplt].set_xlabel(xlabel)
            if vplt[0][:3] == "eta":  # list of log variables
                axes[iplt].set_xscale("log")
            axes[iplt].set_xlim(left=conf.plot.vmin, right=conf.plot.vmax)
            if ivar:
                axes[iplt].legend()
        if conf.rprof.depth:
            axes[0].invert_yaxis()
        ylabel = "Depth" if conf.rprof.depth else "Radius"
        axes[0].set_ylabel(ylabel)
        _helpers.saveplot(conf, fig, fname + stepstr)


def plot_grid(step: Step, conf: Config | None = None) -> None:
    """Plot cell position and thickness.

    The figure is call grid_N.pdf where N is replace by the step index.

    Args:
        step: a `Step` of a `StagyyData` instance.
        conf: configuration.
    """
    if conf is None:
        conf = Config.default_()
    drprof = step.rprofs["dr"]
    fig, (ax1, ax2) = plt.subplots(2, sharex=True)
    ax1.plot(drprof.rad, "-ko")
    ax1.set_ylabel("$r$")
    ax2.plot(drprof.values, "-ko")
    ax2.set_ylabel("$dr$")
    ax2.set_xlim([-0.5, len(drprof.rad) - 0.5])
    ax2.set_xlabel("Cell number")
    _helpers.saveplot(conf, fig, "grid", step.istep)


def cmd(conf: Config) -> None:
    """Plot radial profiles.

    This is the implementation of the `rprof` subcommand.
    """
    sdat = _sdat_from_conf(conf.core)
    view = _helpers.walk(sdat, conf)

    if conf.rprof.grid:
        for step in view.filter(rprofs=True):
            plot_grid(step, conf)

    if conf.rprof.average:
        plot_rprofs(view.rprofs_averaged, conf.rprof.plot, conf)
    else:
        for step in view.filter(rprofs=True):
            plot_rprofs(step.rprofs, conf.rprof.plot, conf)



"""Parse command line arguments."""

from __future__ import annotations

import importlib.resources as imlr
import typing
from pathlib import Path
from types import MappingProxyType

import matplotlib.pyplot as plt
import matplotlib.style as mpls
from loam.cli import CLIManager, Subcmd

from . import __doc__ as doc_module
from . import (
    _styles,
    commands,
    field,
    plates,
    refstate,
    rprof,
    time_series,
)
from ._helpers import baredoc
from .config import Config

if typing.TYPE_CHECKING:
    from typing import Callable


def _sub(cmd: Callable[[Config], None], *sections: str) -> Subcmd:
    """Build Subcmd instance."""
    return Subcmd(baredoc(cmd), *sections, func=cmd)


def _bare_cmd(conf: Config) -> None:
    """Print help message when no arguments are given."""
    print(doc_module)
    print("Run `stagpy -h` for usage")


def _load_mplstyle(conf: Config) -> None:
    """Try to load conf.plot.mplstyle matplotlib style."""
    for style in conf.plot.mplstyle:
        # try packaged version
        style_file = imlr.files(_styles).joinpath(f"{style}.mplstyle")
        if style_file.is_file():
            with imlr.as_file(style_file) as stfile:
                mpls.use(str(stfile))
                continue
        mpls.use(style)
    if conf.plot.xkcd:
        plt.xkcd()


assert doc_module is not None
SUB_CMDS = MappingProxyType(
    {
        "common_": Subcmd(doc_module, "common", func=_bare_cmd),
        "field": _sub(field.cmd, "core", "plot", "scaling"),
        "rprof": _sub(rprof.cmd, "core", "plot", "scaling"),
        "time": _sub(time_series.cmd, "core", "plot", "scaling"),
        "refstate": _sub(refstate.cmd, "core", "plot"),
        "plates": _sub(plates.cmd, "core", "plot", "scaling"),
        "info": _sub(commands.info_cmd, "core", "scaling"),
        "var": _sub(commands.var_cmd),
        "version": _sub(commands.version_cmd),
        "config": _sub(commands.config_cmd),
    }
)


def parse_args(
    conf: Config, arglist: list[str] | None = None
) -> Callable[[Config], None]:
    """Parse cmd line arguments.

    Args:
        conf: configuration.
        arglist: the list of cmd line arguments. If set to None, the arguments
            are taken from :attr:`sys.argv`.

    Returns:
        the function implementing the sub command to be executed.
    """

    def compl_cmd(conf: Config) -> None:
        if conf.completions.zsh:
            filepath = Path("_stagpy.sh")
            print(f"writing zsh completion file {filepath}")
            climan.zsh_complete(filepath, "stagpy", sourceable=True)
        elif conf.completions.bash:
            filepath = Path("stagpy.sh")
            print(f"writing bash completion file {filepath}")
            climan.bash_complete(filepath, "stagpy")
        else:
            print("please choose a shell, `--help` for available options")

    climan = CLIManager(
        conf,
        **SUB_CMDS,
        completions=Subcmd("generate completion scripts", func=compl_cmd),
    )

    cmd_args = climan.parse_args(arglist)
    sub_cmd = cmd_args.loam_sub_name

    if sub_cmd is None:
        return cmd_args.func

    if conf.common.config:
        commands.config_pp(climan.sections_list(sub_cmd), conf)

    _load_mplstyle(conf)

    return cmd_args.func



"""Define high level structure StagyyData and helper classes.

Note:
    The helper classes are not designed to be instantiated on their own, but
    only as attributes of `StagyyData` instances. Users of this module should
    only instantiate [`StagyyData`][stagpy.stagyydata.StagyyData].
"""

from __future__ import annotations

import typing
from collections import abc
from dataclasses import dataclass, field
from functools import cached_property
from itertools import zip_longest
from pathlib import Path

import numpy as np

from . import _helpers, error, phyvars, stagyyparsers, step
from . import datatypes as dt
from ._caching import FieldCache, StepSnap, StepSnapH5, StepSnapLegacy
from .parfile import StagyyPar
from .stagyyparsers import FieldXmf, TracersXmf
from .step import Step

if typing.TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Sequence
    from os import PathLike
    from typing import Callable, TypeAlias

    from numpy.typing import NDArray
    from pandas import DataFrame, Series

    from .config import Core

    StepIndex: TypeAlias = int | slice


@typing.overload
def _as_view_item(obj: Sequence[StepIndex]) -> Sequence[StepIndex]: ...


@typing.overload
def _as_view_item(obj: slice) -> Sequence[slice]: ...


@typing.overload
def _as_view_item(obj: int) -> None: ...


def _as_view_item(
    obj: Sequence[StepIndex] | slice | int,
) -> Sequence[StepIndex] | Sequence[slice] | None:
    """Return None or a suitable iterable to build a StepsView."""
    try:
        iter(obj)  # type: ignore
        return obj  # type: ignore
    except TypeError:
        pass
    if isinstance(obj, slice):
        return (obj,)
    return None


@dataclass(frozen=True)
class Refstate:
    """Reference state profiles.

    The `StagyyData.refstate` attribute is an instance of this class.
    Reference state profiles are accessed through the attributes of this
    object.
    """

    sdat: StagyyData

    @cached_property
    def _data(self) -> tuple[list[list[DataFrame]], list[DataFrame]]:
        """Read reference state profile."""
        reffile = self.sdat.filename("refstat.dat")
        if self.sdat.hdf5 and not reffile.is_file():
            # check legacy folder as well
            reffile = self.sdat.filename("refstat.dat", force_legacy=True)
        data = stagyyparsers.refstate(reffile)
        if data is None:
            raise error.NoRefstateError(self.sdat)
        return data

    @property
    def systems(self) -> list[list[DataFrame]]:
        """Reference state profiles of phases.

        It is a list of list of :class:`pandas.DataFrame` containing
        the reference profiles associated with each phase in each system.

        Example:
            The temperature profile of the 3rd phase in the 1st
            system is

            ```py
            sdat.refstate.systems[0][2]["T"]
            ```
        """
        return self._data[0]

    @property
    def adiabats(self) -> list[DataFrame]:
        """Adiabatic reference state profiles.

        It is a list of :class:`pandas.DataFrame` containing the reference
        profiles associated with each system.  The last item is the combined
        adiabat.

        Example:
            The adiabatic temperature profile of the 2nd system is

            ```py
            sdat.refstate.adiabats[1]["T"]
            ```

            The combined density profile is

            ```py
            sdat.refstate.adiabats[-1]["rho"]
            ```
        """
        return self._data[1]


@dataclass(frozen=True)
class Tseries:
    """Time series.

    The `StagyyData.tseries` attribute is an instance of this class.

    `Tseries` implements the getitem mechanism.  Keys are series names
    defined in `stagpy.phyvars.TIME[_EXTRA]`.  Items are
    [stagpy.datatypes.Tseries][] instances.
    """

    sdat: StagyyData

    @cached_property
    def _cached_extra(self) -> dict[str, dt.Tseries]:
        return {}

    @cached_property
    def _data(self) -> DataFrame | None:
        timefile = self.sdat.filename("TimeSeries.h5")
        data = stagyyparsers.time_series_h5(timefile)
        if data is not None:
            return data
        timefile = self.sdat.filename("time.dat")
        if self.sdat.hdf5 and not timefile.is_file():
            # check legacy folder as well
            timefile = self.sdat.filename("time.dat", force_legacy=True)
        data = stagyyparsers.time_series(timefile)
        return data

    @property
    def _tseries(self) -> DataFrame:
        if self._data is None:
            raise error.MissingDataError(f"No tseries data in {self.sdat}")
        return self._data

    def __getitem__(self, name: str) -> dt.Tseries:
        if (name_alias := phyvars.TIME_ALIAS.get(name)) is not None:
            name = name_alias

        series: NDArray[np.float64]
        if name in self._tseries.columns:
            series = self._tseries[name].to_numpy(dtype=np.float64)
            time = self.time
            if name in phyvars.TIME:
                meta = phyvars.TIME[name]
            else:
                meta = dt.Vart(name, "", "1")
        elif name in self._cached_extra:
            tseries = self._cached_extra[name]
            series = tseries.values
            time = tseries.time
            meta = tseries.meta
        elif name in phyvars.TIME_EXTRA:
            self._cached_extra[name] = phyvars.TIME_EXTRA[name](self.sdat)
            tseries = self._cached_extra[name]
            series = tseries.values
            time = tseries.time
            meta = tseries.meta
        else:
            raise error.UnknownTimeVarError(name)
        return dt.Tseries(series, time, meta)

    def tslice(
        self, name: str, tstart: float | None = None, tend: float | None = None
    ) -> dt.Tseries:
        """Return a [`Tseries`][stagpy.datatypes.Tseries] between specified times.

        Args:
            name: time variable.
            tstart: starting time. Set to None to start at the beginning of
                available data.
            tend: ending time. Set to None to stop at the end of available
                data.
        """
        series = self[name]
        istart = 0
        iend = len(series.time)
        if tstart is not None:
            istart = _helpers.find_in_sorted_arr(tstart, series.time)
        if tend is not None:
            iend = _helpers.find_in_sorted_arr(tend, series.time, True) + 1
        return dt.Tseries(
            series.values[istart:iend],
            series.time[istart:iend],
            series.meta,
        )

    @property
    def time(self) -> NDArray[np.float64]:
        """Time vector."""
        return self._tseries["time"].to_numpy(dtype=np.float64)

    @property
    def isteps(self) -> NDArray[np.float64]:
        """Step indices.

        This is such that `time[istep]` is at step `isteps[istep]`.
        """
        return self._tseries.index.values

    def at_step(self, istep: int) -> Series[np.float64]:
        """Time series output for a given step."""
        return self._tseries.loc[istep]  # type: ignore


@dataclass(frozen=True)
class RprofsAveraged(step.Rprofs):
    """Radial profiles time-averaged over a [`StepsView`][stagpy.stagyydata.StepsView].

    The [`StepsView.rprofs_averaged`][stagpy.stagyydata.StepsView.rprofs_averaged]
    attribute is an instance of this class.

    The current implementation does not take into account time-changing geometry.
    """

    steps: StepsView

    @cached_property
    def _steps_with_rprofs(self) -> StepsView:
        return self.steps.filter(rprofs=True)

    @cached_property
    def _cached_data(self) -> dict[str, dt.Rprof]:
        return {}

    @cached_property
    def _times(self) -> NDArray[np.float64]:
        return np.fromiter((s.time for s in self._steps_with_rprofs), dtype=float)

    @cached_property
    def _dtimes(self) -> NDArray[np.float64]:
        midpoints = (self._times[:-1] + self._times[1:]) / 2
        return np.diff(midpoints, prepend=self._times[0], append=self._times[-1])

    def __getitem__(self, name: str) -> dt.Rprof:
        if name in self._cached_data:
            return self._cached_data[name]
        integral_prof = sum(
            dtime * s.rprofs[name].values
            for dtime, s in zip(self._dtimes, self._steps_with_rprofs, strict=True)
        )
        self._cached_data[name] = dt.Rprof(
            values=integral_prof / (self._times[-1] - self._times[0]),
            rad=self._first_rprofs[name].rad,
            meta=self._first_rprofs[name].meta,
        )
        return self._cached_data[name]

    @property
    def stepstr(self) -> str:
        return self.steps.stepstr

    @cached_property
    def _first_rprofs(self) -> step.RprofsInstant:
        first_step = next(iter(self._steps_with_rprofs))
        return first_step.rprofs

    @property
    def centers(self) -> NDArray[np.float64]:
        return self._first_rprofs.centers

    @property
    def walls(self) -> NDArray[np.float64]:
        return self._first_rprofs.walls

    @property
    def bounds(self) -> tuple[float, float]:
        return self._first_rprofs.bounds


@dataclass(frozen=True)
class Steps:
    """Collections of time steps.

    The `StagyyData.steps` attribute is an instance of this class.
    Time steps (which are [`Step`][stagpy.step.Step] instances) can be
    accessed with the item accessor:

    ```py
    sdat = StagyyData(Path("path/to/run"))
    sdat.steps[istep]  # Step object of the istep-th time step
    ```

    Slices or tuple of istep and slices of `Steps` object are
    `StepsView` instances that you can iterate and filter:

    ```py
    for step in steps[500:]:
        # iterate through all time steps from the 500-th one
        do_something(step)

    for step in steps[-100:].filter(snap=True):
        # iterate through all snapshots present in the last 100 time steps
        do_something(step)

    for step in steps[0,3,5,-2:]:
        # iterate through steps 0, 3, 5 and the last two
        do_something(step)
    ```
    """

    sdat: StagyyData

    def __repr__(self) -> str:
        return f"{self.sdat!r}.steps"

    @cached_property
    def _data(self) -> dict[int, Step]:
        return {}

    @typing.overload
    def __getitem__(self, istep: int) -> Step: ...

    @typing.overload
    def __getitem__(self, istep: slice | Sequence[StepIndex]) -> StepsView: ...

    def __getitem__(self, istep: int | slice | Sequence[StepIndex]) -> Step | StepsView:
        keys = _as_view_item(istep)
        if keys is not None:
            return StepsView(self, keys)
        try:
            istep = int(istep)  # type: ignore
        except ValueError:
            raise error.InvalidTimestepError(
                self.sdat,
                istep,  # type: ignore
                "Time step should be an integer value",
            )
        if istep < 0:
            istep += len(self)
            if istep < 0:
                istep -= len(self)
                raise error.InvalidTimestepError(
                    self.sdat, istep, f"Last istep is {len(self) - 1}"
                )
        if istep not in self._data:
            self._data[istep] = Step(istep, self.sdat)
        return self._data[istep]

    def __delitem__(self, istep: int | None) -> None:
        if istep is not None and istep in self._data:
            self.sdat._field_cache.evict_istep(istep)
            del self._data[istep]

    @cached_property
    def _len(self) -> int:
        return self.sdat.tseries.isteps[-1] + 1

    def __len__(self) -> int:
        return self._len

    def __iter__(self) -> Iterator[Step]:
        return iter(self[:])

    def at_time(self, time: float, after: bool = False) -> Step:
        """Return step corresponding to a given physical time.

        Args:
            time: the physical time requested.
            after: when False (the default), the returned step is such that its
                time is immediately before the requested physical time. When
                True, the returned step is the next one instead (if it exists,
                otherwise the same step is returned).

        Returns:
            the relevant step.
        """
        itime = _helpers.find_in_sorted_arr(time, self.sdat.tseries.time, after)
        return self[self.sdat.tseries.isteps[itime]]

    def filter(
        self,
        snap: bool = False,
        rprofs: bool = False,
        fields: Iterable[str] | None = None,
        func: Callable[[Step], bool] | None = None,
    ) -> StepsView:
        """Build a `StepsView` with requested filters."""
        return self[:].filter(snap, rprofs, fields, func)


@dataclass(frozen=True)
class Snaps:
    """Collection of snapshots.

    The `StagyyData.snaps` attribute is an instance of this class.
    Snapshots (which are [`Step`][stagpy.step.Step] instances) can be accessed
    with the item accessor:

    ```py
    sdat = StagyyData(Path("path/to/run"))
    sdat.snaps[isnap]  # Step object of the isnap-th snapshot
    ```
    """

    sdat: StagyyData

    def __repr__(self) -> str:
        return f"{self.sdat!r}.snaps"

    @typing.overload
    def __getitem__(self, isnap: int) -> Step: ...

    @typing.overload
    def __getitem__(self, isnap: slice | Sequence[StepIndex]) -> StepsView: ...

    def __getitem__(self, isnap: int | slice | Sequence[StepIndex]) -> Step | StepsView:
        keys = _as_view_item(isnap)
        if keys is not None:
            return StepsView(self, keys).filter(snap=True)
        assert isinstance(isnap, int)
        if isnap < 0:
            isnap += len(self)
        istep = self.sdat._step_snap.istep(isnap=isnap)
        if istep is None:
            raise error.InvalidSnapshotError(self.sdat, isnap, "Invalid snapshot index")
        return self.sdat.steps[istep]

    def __delitem__(self, isnap: int | None) -> None:
        if isnap is not None:
            istep = self.sdat._step_snap.istep(isnap=isnap)
            del self.sdat.steps[istep]

    def __len__(self) -> int:
        length = self.sdat._step_snap.len_snap()
        if length <= 0:
            raise error.NoSnapshotError(self.sdat)
        return length

    def __iter__(self) -> Iterator[Step]:
        return iter(self[:])

    def at_time(self, time: float, after: bool = False) -> Step:
        """Return snap corresponding to a given physical time.

        Args:
            time: the physical time requested.
            after: when False (the default), the returned snap is such that its
                time is immediately before the requested physical time. When
                True, the returned snap is the next one instead (if it exists,
                otherwise the same snap is returned).

        Returns:
            the relevant `Step`.
        """
        # in theory, this could be a valid implementation of Steps.at_time
        # but this isn't safe against missing data...
        igm = 0
        igp = len(self) - 1
        while igp - igm > 1:
            istart = igm + (igp - igm) // 2
            if self[istart].time >= time:
                igp = istart
            else:
                igm = istart
        if self[igp].time > time and not after and igp > 0:
            igp -= 1
        return self[igp]

    def filter(
        self,
        snap: bool = False,
        rprofs: bool = False,
        fields: Iterable[str] | None = None,
        func: Callable[[Step], bool] | None = None,
    ) -> StepsView:
        """Build a `StepsView` with requested filters."""
        return self[:].filter(snap, rprofs, fields, func)


@dataclass(frozen=True)
class Filters:
    """Filters on a step view."""

    snap: bool = False
    rprofs: bool = False
    fields: set[str] = field(default_factory=set)
    funcs: list[Callable[[Step], bool]] = field(default_factory=list)

    def passes(self, step: Step) -> bool:
        """Whether a given Step passes the filters."""
        if self.snap and step.isnap is None:
            return False
        if self.rprofs:
            try:
                _ = step.rprofs.centers
            except error.MissingDataError:
                return False
        if any(fld not in step.fields for fld in self.fields):
            return False
        return all(func(step) for func in self.funcs)

    def compose_with(self, other: Filters) -> Filters:
        return Filters(
            snap=self.snap or other.snap,
            rprofs=self.rprofs or other.rprofs,
            fields=self.fields | other.fields,
            funcs=self.funcs + other.funcs,
        )


@dataclass(frozen=True)
class StepsView:
    """Filtered iterator over steps or snaps.

    Instances of this class are returned when taking slices of
    `StagyyData.steps` or `StagyyData.snaps` attributes.

    Args:
        over: steps collection, i.e. `StagyyData.steps` or
            `StagyyData.snaps` attributes.
        items: iterable of isteps/isnaps or slices.
    """

    over: Steps | Snaps
    items: Sequence[StepIndex]
    filters: Filters = field(default_factory=Filters)

    @cached_property
    def rprofs_averaged(self) -> RprofsAveraged:
        """Time-averaged radial profiles."""
        return RprofsAveraged(self)

    @cached_property
    def stepstr(self) -> str:
        """String representation of the requested set of steps."""
        items = []
        no_slice = True
        for item in self.items:
            if isinstance(item, slice):
                items.append("{}:{}:{}".format(*item.indices(len(self.over))))
                no_slice = False
            else:
                items.append(repr(item))
        item_str = ",".join(items)
        if no_slice and len(items) == 1:
            item_str += ","
        colstr = repr(self.over).rsplit(".", maxsplit=1)[-1]
        return f"{colstr}[{item_str}]"

    def _pass(self, item: int) -> bool:
        """Check whether an item passes the filters."""
        try:
            step = self.over[item]
        except (error.InvalidTimestepError, error.InvalidSnapshotError):
            return False
        return self.filters.passes(step)

    def filter(
        self,
        snap: bool = False,
        rprofs: bool = False,
        fields: Iterable[str] | None = None,
        func: Callable[[Step], bool] | None = None,
    ) -> StepsView:
        """Add filters to the view.

        Note that filters are only resolved when the view is iterated.
        Successive calls to :meth:`filter` compose.  For example, with this
        code:

        ```py
        view = sdat.steps[500:].filter(rprofs=True, fields=["T"])
        view.filter(fields=["eta"])
        ```

        the produced `view`, when iterated, will generate the steps after the
        500-th that have radial profiles, and both the temperature and
        viscosity fields.

        Args:
            snap: if true, the step must be a snapshot to pass.
            rprofs: if true, the step must have rprofs data to pass.
            fields: list of fields that must be present to pass.
            func: arbitrary function returning whether a step should pass the
                filter.

        Returns:
            self.
        """
        new_filters = Filters(
            snap=snap,
            rprofs=rprofs,
            fields=set() if fields is None else set(fields),
            funcs=[] if func is None else [func],
        )
        return StepsView(
            over=self.over,
            items=self.items,
            filters=self.filters.compose_with(new_filters),
        )

    def __iter__(self) -> Iterator[Step]:
        for item in self.items:
            if isinstance(item, slice):
                idx = item.indices(len(self.over))
                yield from (self.over[i] for i in range(*idx) if self._pass(i))
            elif self._pass(item):
                yield self.over[item]

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, abc.Iterable):
            return NotImplemented
        return all(s1 is s2 for s1, s2 in zip_longest(self, other))


def _sdat_from_conf(core: Core) -> StagyyData:
    return StagyyData(core.path, core.read_parameters_dat)


@dataclass(frozen=True)
class StagyyData:
    """Generic lazy interface to StagYY output data.

    Args:
        path_hint: path of the StagYY run. It can either be the path of the
            directory containing the par file, or the path of the par file. If
            the path given is a directory, the path of the par file is assumed
            to be path/par.
        read_parameters_dat: read `parameters.dat` file produced by StagYY. This
            flag can be switched off to ignore this file. This is intended for
            runs of StagYY that predate version 1.2.6 for which the
            `parameters.dat` file contained some values affected by internal
            logic.
    """

    path_hint: PathLike[str] | str
    read_parameters_dat: bool = True

    @property
    def path(self) -> Path:
        """Path of StagYY run directory."""
        return self.parpath.parent

    @cached_property
    def parpath(self) -> Path:
        """Path of par file."""
        parpath = Path(self.path_hint)
        if parpath.is_file():
            return parpath
        return parpath / "par"

    @cached_property
    def hdf5(self) -> Path | None:
        """Path of output hdf5 folder if relevant, None otherwise."""
        h5xmf = self.par.h5_output("Data.xmf")
        return h5xmf.parent if h5xmf.is_file() else None

    @cached_property
    def steps(self) -> Steps:
        """Collection of time steps."""
        return Steps(self)

    @cached_property
    def snaps(self) -> Snaps:
        """Collection of snapshots."""
        return Snaps(self)

    @cached_property
    def tseries(self) -> Tseries:
        """Time series data."""
        return Tseries(self)

    @cached_property
    def refstate(self) -> Refstate:
        """Reference state profiles."""
        return Refstate(self)

    @cached_property
    def _dataxmf(self) -> FieldXmf:
        assert self.hdf5 is not None
        return FieldXmf(
            path=self.hdf5 / "Data.xmf",
        )

    @cached_property
    def _topxmf(self) -> FieldXmf:
        assert self.hdf5 is not None
        return FieldXmf(
            path=self.hdf5 / "DataSurface.xmf",
        )

    @cached_property
    def _botxmf(self) -> FieldXmf:
        assert self.hdf5 is not None
        return FieldXmf(
            path=self.hdf5 / "DataBottom.xmf",
        )

    @cached_property
    def _traxmf(self) -> TracersXmf:
        assert self.hdf5 is not None
        return TracersXmf(
            path=self.hdf5 / "DataTracers.xmf",
        )

    @cached_property
    def par(self) -> StagyyPar:
        """Content of par file."""
        return StagyyPar.from_main_par(self.parpath, self.read_parameters_dat)

    @cached_property
    def _rprof_and_times(self) -> tuple[dict[int, DataFrame], DataFrame | None]:
        rproffile = self.filename("rprof.h5")
        data = stagyyparsers.rprof_h5(rproffile)
        if data[1] is not None:
            return data
        rproffile = self.filename("rprof.dat")
        if self.hdf5 and not rproffile.is_file():
            # check legacy folder as well
            rproffile = self.filename("rprof.dat", force_legacy=True)
        return stagyyparsers.rprof(rproffile)

    @property
    def rtimes(self) -> DataFrame | None:
        """Radial profiles times."""
        return self._rprof_and_times[1]

    @cached_property
    def _files(self) -> set[Path]:
        """Set of found binary files output by StagYY."""
        out_dir = self.par.legacy_output("_").parent
        if out_dir.is_dir():
            return set(out_dir.iterdir())
        return set()

    def set_nfields_max(self, nfields: int | None) -> None:
        """Adjust maximum number of scalar fields kept in memory.

        Setting this to a value lower or equal to 5 raises a
        [stagpy.error.InvalidNfieldsError][].  Set this to `None` if
        you do not want any limit on the number of scalar fields kept in
        memory.  Defaults to 50.
        """
        if nfields is not None and nfields <= 5:
            raise error.InvalidNfieldsError(nfields)
        self._field_cache.resize(nfields)

    def filename(
        self,
        fname: str,
        timestep: int | None = None,
        suffix: str = "",
        force_legacy: bool = False,
    ) -> Path:
        """Return name of StagYY output file.

        Args:
            fname: name stem.
            timestep: snapshot number if relevant.
            suffix: optional suffix of file name.
            force_legacy: force returning the legacy output path.

        Returns:
            the path of the output file constructed with the provided segments.
        """
        if timestep is not None:
            fname += f"{timestep:05d}"
        fname += suffix
        if not force_legacy and self.hdf5:
            fpath = self.par.h5_output(fname)
        else:
            fpath = self.par.legacy_output(f"_{fname}")
        return fpath

    def _binfiles_set(self, isnap: int) -> set[Path]:
        """Set of existing binary files at a given snap.

        Args:
            isnap: snapshot index.

        Returns:
            the set of output files available for this snapshot number.
        """
        possible_files = set(
            self.filename(fstem, isnap, force_legacy=True)
            for fstem in phyvars.FIELD_FILES
        )
        return possible_files & self._files

    @cached_property
    def _field_cache(self) -> FieldCache:
        return FieldCache(maxsize=50)

    @cached_property
    def _step_snap(self) -> StepSnap:
        if self.hdf5 is not None:
            return StepSnapH5(sdat=self)
        return StepSnapLegacy(sdat=self)



"""The stagpy module is callable."""

from __future__ import annotations

import os
import signal
import sys
import warnings
from typing import NoReturn


def sigint_handler(*_: object) -> NoReturn:
    """Handler of SIGINT signal.

    It is set when you use StagPy as a command line tool to handle gracefully
    keyboard interruption.
    """
    print("\nSo long, and thanks for all the fish.")
    sys.exit()


def main() -> None:
    """Implement StagPy entry point."""
    debug = os.getenv("STAGPY_DEBUG") is not None
    if debug:
        print(
            "env variable 'STAGPY_DEBUG' is set: StagPy runs in DEBUG mode",
            end="\n\n",
        )
    else:
        signal.signal(signal.SIGINT, sigint_handler)
        warnings.simplefilter("ignore")

    from . import cli, config, error

    conf = config.Config.default_()
    if config.CONFIG_LOCAL.is_file():
        conf.update_from_file_(config.CONFIG_LOCAL)

    try:
        cli.parse_args(conf)(conf)
    except error.StagpyError as err:
        if debug:
            raise
        errtype = type(err).__name__
        print(
            "Oops! StagPy encountered the following problem while processing your request.",
            "Please check the path to your simulation and the command line arguments.",
            "",
            f"{errtype}: {err}",
            sep="\n",
            file=sys.stderr,
        )
        sys.exit()


if __name__ == "__main__":
    main()



from __future__ import annotations

import re
import typing
from abc import ABC, abstractmethod
from collections import deque
from dataclasses import dataclass
from functools import cached_property

from . import phyvars, stagyyparsers
from .error import InvalidSnapshotError

if typing.TYPE_CHECKING:
    from collections.abc import Mapping

    from .datatypes import Field
    from .stagyydata import StagyyData


@dataclass(frozen=True)
class FieldCache:
    """FIFO cache of [Field][]s.

    If `maxsize` is None, entries are never evicted from the cache.
    """

    maxsize: int | None

    @cached_property
    def _stack(self) -> deque[tuple[int, str]]:
        return deque()

    @cached_property
    def _data(self) -> dict[tuple[int, str], Field]:
        return {}

    def _prune(self) -> None:
        if self.maxsize is None:
            return
        while len(self._stack) > self.maxsize:
            elt = self._stack.popleft()
            del self._data[elt]

    def resize(self, new_size: int | None) -> None:
        object.__setattr__(self, "maxsize", new_size)
        self._prune()

    def insert(self, istep: int, name: str, field: Field) -> None:
        key = (istep, name)
        if key not in self._data:
            self._stack.append(key)
        self._data[key] = field
        self._prune()

    def get(self, istep: int, name: str) -> Field | None:
        return self._data.get((istep, name))

    def evict_istep(self, istep: int) -> None:
        to_keep = []
        for key in self._stack:
            if key[0] == istep:
                del self._data[key]
            else:
                to_keep.append(key)
        self._stack.clear()
        self._stack.extend(to_keep)
        assert len(self._stack) == len(self._data)


class StepSnap(ABC):
    """Keep track of the step/snap correspondence."""

    @abstractmethod
    def istep(self, *, isnap: int) -> int | None: ...

    @abstractmethod
    def isnap(self, *, istep: int) -> int | None: ...

    @abstractmethod
    def len_snap(self) -> int: ...


@dataclass(frozen=True)
class StepSnapInfo:
    step_to_snap: Mapping[int, int]
    snap_to_step: Mapping[int, int]
    isnap_max: int


@dataclass(frozen=True)
class StepSnapH5(StepSnap):
    sdat: StagyyData

    @cached_property
    def _info(self) -> StepSnapInfo:
        assert self.sdat.hdf5 is not None
        isnap = -1
        step_to_snap = {}
        snap_to_step = {}
        for isnap, istep in stagyyparsers.read_time_h5(self.sdat.hdf5):
            step_to_snap[istep] = isnap
            snap_to_step[isnap] = istep
        return StepSnapInfo(
            step_to_snap=step_to_snap,
            snap_to_step=snap_to_step,
            isnap_max=isnap,
        )

    def istep(self, *, isnap: int) -> int | None:
        return self._info.snap_to_step.get(isnap)

    def isnap(self, *, istep: int) -> int | None:
        return self._info.step_to_snap.get(istep)

    def len_snap(self) -> int:
        return self._info.isnap_max + 1


@dataclass(frozen=True)
class StepSnapLegacy(StepSnap):
    sdat: StagyyData

    @cached_property
    def _step_to_snap(self) -> dict[int, int | None]:
        return {}

    @cached_property
    def _snap_to_step(self) -> dict[int, int | None]:
        return {}

    @cached_property
    def isnap_max(self) -> int:
        imax = -1
        out_stem = re.escape(self.sdat.par.legacy_output("_").name[:-1])
        rgx = re.compile(f"^{out_stem}_([a-zA-Z]+)([0-9]{{5}})$")
        fstems = set(fstem for fstem in phyvars.FIELD_FILES)
        for fname in self.sdat._files:
            match = rgx.match(fname.name)
            if match is not None and match.group(1) in fstems:
                imax = max(int(match.group(2)), imax)
        return imax

    def len_snap(self) -> int:
        return self.isnap_max + 1

    def istep(self, *, isnap: int) -> int | None:
        if isnap < 0 or isnap > self.isnap_max:
            return None
        istep = self._snap_to_step.get(isnap, -1)
        if istep == -1:
            binfiles = self.sdat._binfiles_set(isnap)
            if binfiles:
                istep = stagyyparsers.field_istep(binfiles.pop())
            else:
                istep = None
            self._snap_to_step[isnap] = istep
            if istep is not None:
                self._step_to_snap[istep] = isnap
        return istep

    def isnap(self, *, istep: int) -> int | None:
        if istep < 0:
            return None
        isnap = self._step_to_snap.get(istep, -1)
        if isnap == -1:
            istep_try = None
            # might be more efficient to do 0 and -1 then bisection, even if
            # that means losing intermediate information
            while (istep_try is None or istep_try < istep) and isnap < 99999:
                isnap += 1
                try:
                    istep_try = self.sdat.snaps[isnap].istep
                except InvalidSnapshotError:
                    pass
                # all intermediate istep could have their isnap to None
                self._snap_to_step[isnap] = istep_try
                if istep_try is not None:
                    self._step_to_snap[istep_try] = isnap

            if istep_try != istep:
                self._step_to_snap[istep] = None

        return self._step_to_snap[istep]



"""Exceptions raised by StagPy."""

from __future__ import annotations

import typing
from dataclasses import dataclass

if typing.TYPE_CHECKING:
    from pathlib import Path

    from .stagyydata import StagyyData
    from .step import Step


class StagpyError(Exception):
    """Base class for exceptions raised by StagPy.

    Note:
        All exceptions derive from this class. To catch any error that might be
        raised by StagPy due to invalid requests/missing data, you only need to
        catch this exception.
    """


@dataclass
class NoSnapshotError(StagpyError):
    """Raised when no snapshot can be found."""

    sdat: StagyyData


@dataclass
class NoGeomError(StagpyError):
    """Raised when no geometry info can be found."""

    step: Step


@dataclass
class NoTimeError(StagpyError):
    """Raised when no time can be found for a step."""

    step: Step


@dataclass
class NoRefstateError(StagpyError):
    """Raised when no refstate output can be found."""

    sdat: StagyyData


@dataclass
class NoParFileError(StagpyError):
    """Raised when no par file can be found."""

    parfile: Path


class NotAvailableError(StagpyError):
    """Raised when a feature is not available yet."""


@dataclass
class ParsingError(StagpyError):
    """Raised when a parsing error occurs."""

    file: Path
    msg: str


@dataclass
class InvalidTimestepError(StagpyError):
    """Raised when invalid time step is requested."""

    sdat: StagyyData
    istep: int
    msg: str


@dataclass
class InvalidSnapshotError(StagpyError):
    """Raised when invalid snapshot is requested."""

    sdat: StagyyData
    isnap: int
    msg: str


@dataclass
class InvalidTimeFractionError(StagpyError):
    """Raised when invalid fraction of series is requested, should be in (0, 1]."""

    fraction: float


@dataclass
class InvalidNfieldsError(StagpyError):
    """Raised when invalid nfields_max is requested."""

    nfields: int


@dataclass
class InvalidZoomError(StagpyError):
    """Raised when invalid zoom is requested, should be in [0, 360]."""

    zoom: float


class MissingDataError(StagpyError):
    """Raised when requested data is not present in output."""


@dataclass
class UnknownVarError(StagpyError):
    """Raised when invalid var is requested."""

    varname: str


class UnknownFieldVarError(UnknownVarError):
    """Raised when invalid field var is requested."""


class UnknownRprofVarError(UnknownVarError):
    """Raised when invalid rprof var is requested."""


class UnknownTimeVarError(UnknownVarError):
    """Raised when invalid time var is requested."""



"""Plot reference state profiles."""

from __future__ import annotations

import matplotlib.pyplot as plt

from . import _helpers
from .config import Config
from .phyvars import REFSTATE
from .stagyydata import StagyyData, _sdat_from_conf


def plot_ref(sdat: StagyyData, var: str, conf: Config | None = None) -> None:
    """Plot one reference state.

    Args:
        sdat: a `StagyyData` instance.
        var: refstate variable, a key of :data:`~stagpy.phyvars.REFSTATE`.
        conf: configuration.
    """
    if conf is None:
        conf = Config.default_()
    fig, axis = plt.subplots()
    adbts = sdat.refstate.adiabats
    if len(adbts) > 2:
        for iad, adia in enumerate(adbts[:-1], 1):
            axis.plot(adia[var], adia["z"], conf.refstate.style, label=f"System {iad}")
    axis.plot(
        adbts[-1][var],
        adbts[-1]["z"],
        conf.refstate.style,
        color="k",
        label="Combined profile",
    )
    if var == "Tcond":
        axis.set_xscale("log")
    axis.set_xlabel(REFSTATE[var].description)
    axis.set_ylabel("z Position")
    if len(adbts) > 2:
        axis.legend()
    _helpers.saveplot(conf, fig, f"refstate_{var}")


def cmd(conf: Config) -> None:
    """Plot reference state profiles.

    This is the implementation of the `refstate` subcommand.
    """
    sdat = _sdat_from_conf(conf.core)

    for var in conf.refstate.plot:
        plot_ref(sdat, var, conf)



"""Plots time series."""

from __future__ import annotations

import typing

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.integrate import trapezoid

from . import _helpers
from .config import Config
from .error import InvalidTimeFractionError
from .stagyydata import StagyyData, _sdat_from_conf

if typing.TYPE_CHECKING:
    from collections.abc import Sequence

    from pandas import DataFrame


def _collect_marks(sdat: StagyyData, conf: Config) -> list[float]:
    """Concatenate `mark*` config variable."""
    times = list(conf.time.marktimes)
    times.extend(step.timeinfo["time"] for step in sdat.snaps[conf.time.marksnaps])
    times.extend(step.timeinfo["time"] for step in sdat.steps[conf.time.marksteps])
    return times


def plot_time_series(
    sdat: StagyyData,
    names: Sequence[Sequence[Sequence[str]]],
    conf: Config | None = None,
) -> None:
    """Plot requested time series.

    Args:
        sdat: a `StagyyData` instance.
        names: time series names organized by figures, plots and subplots.
        conf: configuration.
    """
    if conf is None:
        conf = Config.default_()
    time_marks = _collect_marks(sdat, conf)
    for vfig in names:
        tstart = conf.time.tstart
        tend = conf.time.tend
        fig, axes = plt.subplots(
            nrows=len(vfig), sharex=True, figsize=(12, 2 * len(vfig))
        )
        axes = [axes] if len(vfig) == 1 else axes
        fname = ["time"]
        for iplt, vplt in enumerate(vfig):
            ylabel = None
            series_on_plt = [
                sdat.tseries.tslice(tvar, conf.time.tstart, conf.time.tend)
                for tvar in vplt
            ]
            ptstart = min(series.time[0] for series in series_on_plt)
            ptend = max(series.time[-1] for series in series_on_plt)
            tstart = ptstart if tstart is None else min(ptstart, tstart)
            tend = ptend if tend is None else max(ptend, tend)
            fname.extend(vplt)
            for tseries in series_on_plt:
                axes[iplt].plot(
                    tseries.time,
                    tseries.values,
                    conf.time.style,
                    label=tseries.meta.description,
                )
                lbl = tseries.meta.kind
                if ylabel is None:
                    ylabel = lbl
                elif ylabel != lbl:
                    ylabel = ""
            if len(series_on_plt) == 1:
                ylabel = series_on_plt[0].meta.description
            if ylabel:
                axes[iplt].set_ylabel(ylabel)
            if vplt[0][:3] == "eta":  # list of log variables
                axes[iplt].set_yscale("log")
            axes[iplt].set_ylim(bottom=conf.plot.vmin, top=conf.plot.vmax)
            if len(series_on_plt) > 1:
                axes[iplt].legend()
            axes[iplt].tick_params()
            for time_mark in time_marks:
                axes[iplt].axvline(time_mark, color="black", linestyle="--")
        axes[-1].set_xlabel("Time")
        axes[-1].set_xlim(tstart, tend)
        axes[-1].tick_params()
        _helpers.saveplot(conf, fig, "_".join(fname))


def compstat(
    sdat: StagyyData,
    *names: str,
    tstart: float | None = None,
    tend: float | None = None,
) -> DataFrame:
    """Compute statistics from series output by StagYY.

    Args:
        sdat: a `StagyyData` instance.
        names: variables whose statistics should be computed.
        tstart: starting time. Set to None to start at the beginning of
            available data.
        tend: ending time. Set to None to stop at the end of available data.

    Returns:
        statistics. `"mean"` and `"rms"` as index, variable names as columns.
    """
    stats = pd.DataFrame(columns=names, index=["mean", "rms"])
    for name in names:
        series = sdat.tseries.tslice(name, tstart, tend)
        delta_time = series.time[-1] - series.time[0]
        mean = trapezoid(series.values, x=series.time) / delta_time
        stats.loc["mean", name] = mean
        stats.loc["rms", name] = np.sqrt(
            trapezoid((series.values - mean) ** 2, x=series.time) / delta_time
        )
    return stats


def cmd(conf: Config) -> None:
    """Plot time series.

    This is the implementation of the `time` subcommand.
    """
    sdat = _sdat_from_conf(conf.core)
    if conf.time.fraction is not None:
        if not 0 < conf.time.fraction <= 1:
            raise InvalidTimeFractionError(conf.time.fraction)
        conf.time.tend = None
        t_0 = sdat.tseries.time[0]
        t_f = sdat.tseries.time[-1]
        conf.time.tstart = t_0 * conf.time.fraction + t_f * (1 - conf.time.fraction)

    plot_time_series(sdat, conf.time.plot, conf)

    if conf.time.compstat:
        stats = compstat(
            sdat, *conf.time.compstat, tstart=conf.time.tstart, tend=conf.time.tend
        )
        stats.to_csv("statistics.dat")



"""Various helper functions and classes."""

from __future__ import annotations

import typing
from inspect import getdoc

import matplotlib.pyplot as plt
import numpy as np

if typing.TYPE_CHECKING:
    from matplotlib.figure import Figure
    from numpy.typing import NDArray

    from .config import Config
    from .stagyydata import StagyyData, StepsView


def walk(sdat: StagyyData, conf: Config) -> StepsView:
    """Return view on configured steps slice."""
    if conf.core.timesteps:
        return sdat.steps[conf.core.timesteps]
    return sdat.snaps[conf.core.snapshots]


def out_name(conf: Config, stem: str, timestep: int | None = None) -> str:
    """Return StagPy out file name.

    Args:
        stem: short description of file content.
        timestep: timestep if relevant.

    Returns:
        the output file name.
    """
    if conf.core.shortname:
        return conf.core.outname
    if timestep is not None:
        stem = f"{stem}{timestep:05d}"
    return conf.core.outname + "_" + stem


def scilabel(value: float, precision: int = 2) -> str:
    """Build scientific notation of some value.

    This is dedicated to use in labels displaying scientific values.

    Args:
        value: numeric value to format.
        precision: number of decimal digits.

    Returns:
        the scientific notation of the specified value.
    """
    man, exps = f"{value:.{precision}e}".split("e")
    exp = int(exps)
    return rf"{man}\times 10^{{{exp}}}"


def saveplot(
    conf: Config,
    fig: Figure,
    stem: str,
    timestep: int | None = None,
    close: bool = True,
) -> None:
    """Save matplotlib figure.

    You need to provide :data:`stem` as a positional or keyword argument (see
    :func:`out_name`).

    Args:
        fig: the :class:`matplotlib.figure.Figure` to save.
        stem: short description of file content.
        timestep: timestep if relevant.
        close: whether to close the figure.
    """
    oname = out_name(conf, stem, timestep)
    fig.savefig(
        f"{oname}.{conf.plot.format}", format=conf.plot.format, bbox_inches="tight"
    )
    if close:
        plt.close(fig)


def baredoc(obj: object) -> str:
    """Return the first line of the docstring of an object.

    Trailing periods and spaces as well as leading spaces are removed from the
    output.

    Args:
        obj: any Python object.

    Returns:
        str: the first line of the docstring of obj.
    """
    doc = getdoc(obj)
    if not doc:
        return ""
    doc = doc.splitlines()[0]
    return doc.rstrip(" .").lstrip()


def find_in_sorted_arr(
    value: float, array: NDArray[np.float64], after: bool = False
) -> int:
    """Return position of element in a sorted array.

    Returns:
        the maximum position i such as array[i] <= value.  If after is True, it
        returns the min i such as value <= array[i] (or 0 if such an index does
        not exist).
    """
    ielt = array.searchsorted(value)
    if ielt == array.size:
        ielt -= 1
    if not after and array[ielt] != value and ielt > 0:
        ielt -= 1
    return int(ielt)



"""Parsers of StagYY output files.

Note:
    These functions are low level utilities. You should not use these unless
    you know what you are doing. To access StagYY output data, use an instance
    of [`StagyyData`][stagpy.stagyydata.StagyyData].
"""

from __future__ import annotations

import re
import typing
from dataclasses import dataclass
from functools import cached_property
from itertools import product
from operator import itemgetter

import h5py
import numpy as np
import pandas as pd

from .error import ParsingError
from .phyvars import FIELD_FILES_H5, RPROF, SFIELD_FILES_H5
from .xdmf import XmlStream

if typing.TYPE_CHECKING:
    from collections.abc import Iterator, Mapping
    from pathlib import Path
    from typing import Any, BinaryIO
    from xml.etree.ElementTree import Element

    from numpy.typing import NDArray
    from pandas import DataFrame


def _resize(names: list[str], nnames: int) -> None:
    """Truncate or extend names so that its len is nnames.

    The list is modified in-place.

    Args:
        names: list of names.
        nnames: desired number of names.
    """
    names.extend(map(str, range(nnames - len(names))))
    del names[nnames:]


def time_series(timefile: Path) -> DataFrame | None:
    """Read temporal series text file.

    Args:
        timefile: path of the time.dat file.

    Returns:
        A `pandas.DataFrame` containing the time series, organized by
            variables in columns and time steps in rows.
    """
    if not timefile.is_file():
        return None

    with timefile.open() as fid:
        colnames = fid.readline().strip().split()
    # extra columns in case some were added mid-run
    _resize(colnames, len(colnames) + 10)

    data = pd.read_csv(
        timefile,
        sep=r"\s+",
        dtype=str,
        header=None,
        names=colnames,
        skiprows=1,
        index_col="istep",
        engine="c",
        memory_map=True,
        on_bad_lines="skip",
    )
    data = data.apply(pd.to_numeric, raw=True, errors="coerce")

    # detect useless lines produced when run is restarted
    rows_to_del = []
    irow = len(data) - 1
    while irow > 0:
        iprev = irow - 1
        while iprev >= 0 and data.index[irow] <= data.index[iprev]:
            rows_to_del.append(iprev)
            iprev -= 1
        irow = iprev
    if rows_to_del:
        rows_to_keep = set(range(len(data))) - set(rows_to_del)
        data = data.take(list(rows_to_keep))
    data.dropna(axis="columns", how="all", inplace=True)

    return data


def time_series_h5(timefile: Path) -> DataFrame | None:
    """Read temporal series HDF5 file.

    Args:
        timefile: path of the TimeSeries.h5 file.

    Returns:
        A `pandas.DataFrame` containing the time series, organized by
            variables in columns and the time steps in rows.
    """
    if not timefile.is_file():
        return None
    with h5py.File(timefile, "r") as h5f:
        dset = h5f["tseries"]
        _, ncols = dset.shape
        ncols -= 1  # first is istep
        colnames = list(h5f["names"].asstr()[()])
        _resize(colnames, ncols + 1)
        data = dset[()]
    pdf = pd.DataFrame(
        data[:, 1:],
        index=data[:, 0].astype(np.int64),
        columns=colnames[1:],
    )
    # remove duplicated lines in case of restart
    return pdf.loc[~pdf.index.duplicated(keep="last")]


def _extract_rsnap_isteps(
    rproffile: Path, data: DataFrame
) -> list[tuple[int, float, DataFrame]]:
    """Extract istep, time and build separate rprof df."""
    step_regex = re.compile(r"^\*+step:\s*(\d+) ; time =\s*(\S+)")
    isteps = []  # list of (istep, time, df)
    line = " "
    with rproffile.open() as stream:
        while line[0] != "*":
            line = stream.readline()
        match = step_regex.match(line)
        if match is None:
            raise ParsingError(rproffile, f"Badly formatted line {line!r}")
        istep = int(match.group(1))
        time = float(match.group(2))
        nlines = 0
        iline = 0
        for line in stream:
            if line[0] == "*":
                isteps.append((istep, time, data.iloc[iline - nlines : iline]))
                match = step_regex.match(line)
                if match is None:
                    raise ParsingError(rproffile, f"Badly formatted line {line!r}")
                istep = int(match.group(1))
                time = float(match.group(2))
                nlines = 0
                # remove useless lines produced when run is restarted
                while isteps and istep <= isteps[-1][0]:
                    isteps.pop()
            else:
                nlines += 1
                iline += 1
        isteps.append((istep, time, data.iloc[iline - nlines : iline]))
    return isteps


def rprof(rproffile: Path) -> tuple[dict[int, DataFrame], DataFrame | None]:
    """Extract radial profiles data.

    Args:
        rproffile: path of the rprof.dat file.

    Returns:
        profs: a dict mapping istep to radial profiles.
        times: the time indexed by time steps.
    """
    if not rproffile.is_file():
        return {}, None

    with rproffile.open() as fid:
        colnames = fid.readline().strip().split()
    if not colnames:
        colnames = list(RPROF.keys())

    data = pd.read_csv(
        rproffile,
        sep=r"\s+",
        dtype=str,
        header=None,
        comment="*",
        skiprows=1,
        engine="c",
        memory_map=True,
        on_bad_lines="skip",
    )
    data = data.apply(pd.to_numeric, raw=True, errors="coerce")

    isteps = _extract_rsnap_isteps(rproffile, data)

    all_data = {}
    for istep, _, step_df in isteps:
        step_df.index = pd.RangeIndex(step_df.shape[0])  # check whether necessary
        step_cols = list(colnames)
        _resize(step_cols, step_df.shape[1])
        step_df.columns = pd.Index(step_cols)
        all_data[istep] = step_df

    df_times = pd.DataFrame(
        list(map(itemgetter(1), isteps)), index=pd.Index(map(itemgetter(0), isteps))
    )
    return all_data, df_times


def rprof_h5(rproffile: Path) -> tuple[dict[int, DataFrame], DataFrame | None]:
    """Extract radial profiles data.

    Args:
        rproffile: path of the rprof.h5 file.

    Returns:
        profs: a dict mapping istep to radial profiles.
        times: the time indexed by time steps.
    """
    if not rproffile.is_file():
        return {}, None
    isteps = []
    data = {}
    with h5py.File(rproffile, "r") as h5f:
        dnames = sorted(dname for dname in h5f.keys() if dname.startswith("rprof_"))
        colnames = h5f["names"].asstr()[()]
        for dname in dnames:
            dset = h5f[dname]
            arr = dset[()]
            istep = dset.attrs["istep"]
            step_cols = list(colnames)
            _resize(step_cols, arr.shape[1])  # check shape
            data[istep] = pd.DataFrame(arr, columns=step_cols)
            isteps.append((istep, dset.attrs["time"]))

    df_times = pd.DataFrame(
        list(map(itemgetter(1), isteps)), index=pd.Index(map(itemgetter(0), isteps))
    )
    return data, df_times


def _clean_names_refstate(names: list[str]) -> list[str]:
    """Uniformization of refstate profile names."""
    to_clean = {
        "Tref": "T",
        "rhoref": "rho",
        "tcond": "Tcond",
    }
    return [to_clean.get(n, n) for n in names]


def refstate(
    reffile: Path, ncols: int = 8
) -> tuple[list[list[DataFrame]], list[DataFrame]] | None:
    """Extract reference state profiles.

    Args:
        reffile: path of the refstate file.
        ncols: number of columns.

    Returns:
        syst: list of list of `pandas.DataFrame` containing the reference
            state profiles for each system and each phase in these systems.
        adia: list of `pandas.DataFrame` containing the adiabatic reference
            state profiles for each system, the last item being the combined
            adiabat.
    """
    if not reffile.is_file():
        return None
    data = pd.read_csv(
        reffile,
        sep=r"\s+",
        dtype=str,
        header=None,
        names=range(ncols),
        engine="c",
        memory_map=True,
        on_bad_lines="skip",
    )
    data = data.apply(pd.to_numeric, raw=True, errors="coerce")
    # drop lines corresponding to metadata
    data.dropna(subset=[0], inplace=True)
    isystem = -1
    systems: list[list[list[str]]] = [[]]
    adiabats: list[list[str]] = []
    with reffile.open() as rsf:
        for line in rsf:
            line = line.lstrip()
            if line.startswith("SYSTEM"):
                isystem += 1
                if isystem > 0:
                    systems.append([])
            elif line.startswith("z"):
                systems[isystem].append(_clean_names_refstate(line.split()))
            elif line.startswith("ADIABAT") or line.startswith("COMBINED"):
                line = line.partition(":")[-1]
                adiabats.append(_clean_names_refstate(line.split()))
    nprofs = sum(map(len, systems)) + len(adiabats)
    nzprof = len(data) // nprofs
    iprof = 0
    syst: list[list[DataFrame]] = []
    adia: list[DataFrame] = []
    for isys, layers in enumerate(systems):
        syst.append([])
        for layer in layers:
            ibgn = iprof * nzprof
            iend = ibgn + nzprof
            syst[isys].append(
                pd.DataFrame(data.iloc[ibgn:iend, : len(layer)].values, columns=layer)
            )
            iprof += 1
        if len(layers) > 1:
            ibgn = iprof * nzprof
            iend = ibgn + nzprof
            cols = adiabats.pop(0)
            adia.append(
                pd.DataFrame(data.iloc[ibgn:iend, : len(cols)].values, columns=cols)
            )
            iprof += 1
        else:
            adia.append(syst[isys][0])
    ibgn = iprof * nzprof
    iend = ibgn + nzprof
    cols = adiabats.pop(0)
    adia.append(pd.DataFrame(data.iloc[ibgn:iend, : len(cols)].values, columns=cols))
    return syst, adia


@dataclass(frozen=True)
class _Cursor:
    fid: BinaryIO
    int_type: type[np.integer]
    float_type: type[np.floating]

    def reset_with_64_bits(self) -> _Cursor:
        self.fid.seek(0)
        return _Cursor(
            fid=self.fid,
            int_type=np.int64,
            float_type=np.float64,
        )

    def string(self, nbytes: int) -> str:
        return b"".join(np.fromfile(self.fid, "b", nbytes)).strip().decode()

    def single_int(self) -> np.integer:
        return np.fromfile(self.fid, self.int_type, 1)[0]

    def single_float(self) -> np.floating:
        return np.fromfile(self.fid, self.float_type, 1)[0]

    def ints(self, count: int | np.integer) -> NDArray[np.integer]:
        return np.fromfile(self.fid, self.int_type, count)

    def floats(self, count: int | np.integer) -> NDArray[np.floating]:
        return np.fromfile(self.fid, self.float_type, count)


@dataclass(frozen=True)
class _HeaderInfo:
    """Header information."""

    magic: int
    nval: int
    sfield: bool
    cursor: _Cursor
    header: dict[str, Any]


def _legacy_header(
    filepath: Path, fid: BinaryIO, stop_at_istep: bool = False
) -> _HeaderInfo:
    """Read the header of a legacy binary file."""
    cursor = _Cursor(fid=fid, int_type=np.int32, float_type=np.float32)
    magic = cursor.single_int().item()
    if magic > 8000:  # 64 bits
        cursor = cursor.reset_with_64_bits()
        if magic != cursor.single_int():
            raise ParsingError(filepath, "inconsistent magic number in 64 bits")
        magic -= 8000

    # check nb components
    nval = 1
    sfield = False
    if magic > 400:
        nval = 4
    elif magic > 300:
        nval = 3
    elif magic > 100:
        sfield = True

    magic %= 100
    if magic < 9 or magic > 12:
        raise ParsingError(filepath, f"{magic=:d} not supported")

    header_info = _HeaderInfo(magic, nval, sfield, cursor, {})
    header = header_info.header
    # extra ghost point in horizontal direction
    header["xyp"] = int(nval == 4)  # magic >= 9

    # total number of values in relevant space basis
    # (e1, e2, e3) = (theta, phi, radius) in spherical geometry
    #              = (x, y, z)            in cartesian geometry
    header["nts"] = cursor.ints(3)

    # number of blocks, 2 for yinyang or cubed sphere
    header["ntb"] = cursor.single_int()  # magic >= 7

    # aspect ratio
    header["aspect"] = cursor.floats(2)

    # number of parallel subdomains
    header["ncs"] = cursor.ints(3)  # (e1, e2, e3) space
    header["ncb"] = cursor.single_int()  # magic >= 8, blocks

    # r - coordinates
    # rgeom[0:self.nrtot+1, 0] are edge radial position
    # rgeom[0:self.nrtot, 1] are cell-center radial position
    header["rgeom"] = cursor.floats(header["nts"][2] * 2 + 1)  # magic >= 2
    header["rgeom"] = np.resize(header["rgeom"], (header["nts"][2] + 1, 2))

    header["rcmb"] = cursor.single_float()  # magic >= 7

    header["ti_step"] = cursor.single_int()  # magic >= 3
    if stop_at_istep:
        return header_info

    header["ti_ad"] = cursor.single_float()  # magic >= 3
    header["erupta_total"] = cursor.single_float()  # magic >= 5
    if magic >= 12:
        header["erupta_ttg"] = cursor.single_float()
        header["intruda"] = cursor.floats(2)
        header["ttg_mass"] = cursor.floats(3)
    else:
        header["erupta_ttg"] = 0.0
        header["intruda"] = np.zeros(2)
        header["ttg_mass"] = np.zeros(3)
    header["bot_temp"] = cursor.single_float()  # magic >= 6
    header["core_temp"] = cursor.single_float() if magic >= 10 else 1.0
    header["ocean_mass"] = cursor.single_float() if magic >= 11 else 0.0

    # magic >= 4
    header["e1_coord"] = cursor.floats(header["nts"][0])
    header["e2_coord"] = cursor.floats(header["nts"][1])
    header["e3_coord"] = cursor.floats(header["nts"][2])

    return header_info


def field_istep(fieldfile: Path) -> int | None:
    """Read istep from binary field file.

    Args:
        fieldfile: path of the binary field file.

    Returns:
        the time step at which the binary file was written.
    """
    if not fieldfile.is_file():
        return None
    with fieldfile.open("rb") as fid:
        hdr = _legacy_header(fieldfile, fid, stop_at_istep=True)
    return hdr.header["ti_step"]


def field_header(fieldfile: Path) -> dict[str, Any] | None:
    """Read header info from binary field file.

    Args:
        fieldfile: path of the binary field file.

    Returns:
        the header information of the binary file.
    """
    if not fieldfile.is_file():
        return None
    with fieldfile.open("rb") as fid:
        hdr = _legacy_header(fieldfile, fid)
    return hdr.header


def fields(fieldfile: Path) -> tuple[dict[str, Any], NDArray[np.float64]] | None:
    """Extract fields data.

    Args:
        fieldfile: path of the binary field file.

    Returns:
        the tuple `(header, fields)`. `fields` is an array of scalar fields
            indexed by variable, x-direction, y-direction, z-direction, block.
    """
    if not fieldfile.is_file():
        return None
    with fieldfile.open("rb") as fid:
        hdr = _legacy_header(fieldfile, fid)
        header = hdr.header
        cursor = hdr.cursor

        # READ FIELDS
        # number of points in (e1, e2, e3) directions PER CPU
        npc = header["nts"] // header["ncs"]
        # number of blocks per cpu
        nbk = header["ntb"] // header["ncb"]
        # number of values per 'read' block
        npi = (
            (npc[0] + header["xyp"])
            * (npc[1] + header["xyp"])
            * npc[2]
            * nbk
            * hdr.nval
        )

        header["scalefac"] = cursor.single_float() if hdr.nval > 1 else 1.0

        flds = np.zeros(
            (
                hdr.nval,
                header["nts"][0] + header["xyp"],
                header["nts"][1] + header["xyp"],
                header["nts"][2],
                header["ntb"],
            )
        )

        # loop over parallel subdomains
        for icpu in product(
            range(header["ncb"]),
            range(header["ncs"][2]),
            range(header["ncs"][1]),
            range(header["ncs"][0]),
        ):
            # read the data for one CPU
            data_cpu = cursor.floats(npi) * header["scalefac"]

            # icpu is (icpu block, icpu z, icpu y, icpu x)
            # data from file is transposed to obtained a field
            # array indexed with (x, y, z, block), as in StagYY
            flds[
                :,
                icpu[3] * npc[0] : (icpu[3] + 1) * npc[0] + header["xyp"],  # x
                icpu[2] * npc[1] : (icpu[2] + 1) * npc[1] + header["xyp"],  # y
                icpu[1] * npc[2] : (icpu[1] + 1) * npc[2],  # z
                icpu[0] * nbk : (icpu[0] + 1) * nbk,  # block
            ] = np.transpose(
                data_cpu.reshape(
                    (
                        nbk,
                        npc[2],
                        npc[1] + header["xyp"],
                        npc[0] + header["xyp"],
                        hdr.nval,
                    )
                )
            )
        if hdr.sfield:
            # for surface fields, variables are written along z direction
            flds = np.swapaxes(flds, 0, 3)
    return header, flds


def tracers(tracersfile: Path) -> dict[str, list[NDArray[np.floating]]] | None:
    """Extract tracers data.

    Args:
        tracersfile: path of the binary tracers file.

    Returns:
        Tracers data organized by attribute names and blocks.
    """
    if not tracersfile.is_file():
        return None
    tra: dict[str, list[NDArray[np.floating]]] = {}
    with tracersfile.open("rb") as fid:
        cursor = _Cursor(fid=fid, int_type=np.int32, float_type=np.float32)
        magic = cursor.single_int().item()
        if magic > 8000:  # 64 bits
            cursor = cursor.reset_with_64_bits()
            if magic != cursor.single_int():
                raise ParsingError(tracersfile, "inconsistent magic number in 64 bits")
            magic -= 8000
        if magic < 100:
            raise ParsingError(
                tracersfile, "magic > 100 expected to get tracervar info"
            )
        nblk = magic % 100
        cursor.floats(2)  # aspect ratio
        cursor.single_int()  # istep
        cursor.single_float()  # time
        ninfo = cursor.single_int()
        ntra = cursor.ints(nblk)
        cursor.single_float()  # tracer ideal mass
        curv = cursor.single_int()
        if curv:
            cursor.single_float()  # r_cmb
        infos = []  # list of info names
        for _ in range(ninfo):
            infos.append(cursor.string(16))
            tra[infos[-1]] = []
        if magic > 200:
            ntrace_elt = cursor.single_int()
            if ntrace_elt > 0:
                cursor.floats(ntrace_elt)  # outgassed
        for ntrab in ntra:  # blocks
            data = cursor.floats(ntrab * ninfo)
            for idx, info in enumerate(infos):
                tra[info].append(data[idx::ninfo])
    return tra


def _read_group_h5(filename: Path, groupname: str) -> NDArray[np.float64]:
    """Return group content.

    Args:
        filename: path of hdf5 file.
        groupname: name of group to read.

    Returns:
        content of group.
    """
    try:
        with h5py.File(filename, "r") as h5f:
            data = h5f[groupname][()]
    except OSError as err:
        # h5py doesn't always include the filename in its error messages
        err.args += (filename,)
        raise
    return data  # need to be reshaped


def _make_3d(field: NDArray[np.float64], twod: str | None) -> NDArray[np.float64]:
    """Add a dimension to field if necessary.

    Args:
        field: the field that needs to be 3d.
        twod: 'XZ', 'YZ' or None depending on what is relevant.

    Returns:
        reshaped field.
    """
    shp = list(field.shape)
    if twod and "X" in twod:
        shp.insert(1, 1)
    elif twod:
        shp.insert(0, 1)
    return field.reshape(shp)


def _ncores(
    meshes: list[dict[str, NDArray[np.float64]]], twod: str | None
) -> NDArray[np.float64]:
    """Compute number of nodes in each direction."""
    nnpb = len(meshes)  # number of nodes per block
    nns = [1, 1, 1]  # number of nodes in x, y, z directions
    if twod is None or "X" in twod:
        while (
            nnpb > 1
            and meshes[nns[0]]["X"][0, 0, 0] == meshes[nns[0] - 1]["X"][-1, 0, 0]
        ):
            nns[0] += 1
            nnpb -= 1
    cpu = lambda icy: icy * nns[0]  # noqa: E731
    if twod is None or "Y" in twod:
        while (
            nnpb > 1
            and meshes[cpu(nns[1])]["Y"][0, 0, 0]
            == meshes[cpu(nns[1] - 1)]["Y"][0, -1, 0]
        ):
            nns[1] += 1
            nnpb -= nns[0]
    cpu = lambda icz: icz * nns[0] * nns[1]  # noqa: E731
    while (
        nnpb > 1
        and meshes[cpu(nns[2])]["Z"][0, 0, 0] == meshes[cpu(nns[2] - 1)]["Z"][0, 0, -1]
    ):
        nns[2] += 1
        nnpb -= nns[0] * nns[1]
    return np.array(nns)


def _conglomerate_meshes(
    meshin: list[dict[str, NDArray[np.float64]]], header: dict[str, Any]
) -> dict[str, NDArray[np.float64]]:
    """Conglomerate meshes from several cores into one."""
    meshout = {}
    npc = header["nts"] // header["ncs"]
    shp = [val + 1 if val != 1 else 1 for val in header["nts"]]
    x_p = int(shp[0] != 1)
    y_p = int(shp[1] != 1)
    for coord in meshin[0]:
        meshout[coord] = np.zeros(shp)
    for icore in range(np.prod(header["ncs"])):
        ifs = [
            icore // np.prod(header["ncs"][:i]) % header["ncs"][i] * npc[i]
            for i in range(3)
        ]
        for coord, mesh in meshin[icore].items():
            meshout[coord][
                ifs[0] : ifs[0] + npc[0] + x_p,
                ifs[1] : ifs[1] + npc[1] + y_p,
                ifs[2] : ifs[2] + npc[2] + 1,
            ] = mesh
    return meshout


def _try_text(file: Path, elt: Element) -> str:
    """Try getting text of element or raise a ParsingError."""
    text = elt.text
    if text is None:
        raise ParsingError(file, f"Element {elt} has no 'text'")
    return text


def _ifile_isnap(file: Path, elt: Element) -> tuple[int, int]:
    """Extract ifile and isnap from H5 file name/dataset."""
    data_text = _try_text(file, elt)
    h5file, group = data_text.strip().split(":/", 1)
    isnap = int(group[-5:])
    ifile = int(h5file[-14:-9])
    return ifile, isnap


def _count_subdomains(xs: XmlStream, i0_yin: int) -> tuple[range, range]:
    i1_yin = i0_yin + 1
    i0_yang = 0
    i1_yang = 0
    for _ in xs.iter_tag("Grid"):
        if xs.current.attrib["GridType"] == "Collection":
            break
        if (name := xs.current.attrib["Name"]).startswith("meshYang"):
            if i1_yang == 0:
                i0_yang = int(name[-5:]) - 1
                i1_yang = i0_yang + (i1_yin - i0_yin)
        else:
            i1_yin += 1
        xs.drop()
    return range(i0_yin, i1_yin), range(i0_yang, i1_yang)


@dataclass(frozen=True)
class FieldSub:
    file: Path
    dataset: str
    icore: int
    iblock: int
    shape: tuple[int, ...]


@dataclass(frozen=True)
class XmfEntry:
    isnap: int
    time: float | None
    mo_lambda: float | None
    mo_thick_sol: float | None
    yin_yang: bool
    twod: str | None
    coord_filepattern: str
    coord_shape: tuple[int, ...]
    range_yin: range
    range_yang: range
    fields: Mapping[str, tuple[int, tuple[int, ...]]]

    def coord_files_yin(self, path_root: Path) -> Iterator[Path]:
        for icore in self.range_yin:
            yield path_root / self.coord_filepattern.format(icore=icore + 1)

    def _fsub(self, path_root: Path, name: str, icore: int, yang: bool) -> FieldSub:
        ifile, shape = self.fields[name]
        yy_tag = "2" if yang else "_"
        ic = icore + 1
        return FieldSub(
            file=path_root / f"{name}_{ifile:05d}_{ic:05d}.h5",
            dataset=f"{name}{yy_tag}{ic:05d}_{self.isnap:05d}",
            icore=icore,
            iblock=int(yang),
            shape=shape,
        )

    def field_subdomains(self, path_root: Path, name: str) -> Iterator[FieldSub]:
        if name not in self.fields:
            return
        for icore in self.range_yin:
            yield self._fsub(path_root, name, icore, False)
        for icore in self.range_yang:
            yield self._fsub(path_root, name, icore, True)


@dataclass(frozen=True)
class FieldXmf:
    path: Path

    def _get_dims(self, elt: Element) -> tuple[int, ...]:
        dims = elt.attrib["Dimensions"].split()
        return tuple(map(int, dims))

    @cached_property
    def _data(self) -> Mapping[int, XmfEntry]:
        xs = XmlStream(filepath=self.path)
        data = {}
        for _ in xs.iter_tag("Time"):
            time = float(xs.current.attrib["Value"])
            xs.advance()
            extra: dict[str, float] = {}
            while xs.current.tag != "Grid":
                # mo_lambda, mo_thick_sol
                extra[xs.current.tag] = float(xs.current.attrib["Value"])
                xs.advance()

            mesh_name = xs.current.attrib["Name"]
            yin_yang = mesh_name.startswith("meshYin")
            i0_yin = int(mesh_name[-5:]) - 1
            twod = None

            xs.skip_to_tag("Geometry")
            with xs.load() as elt_geom:
                if elt_geom.get("Type") == "X_Y":
                    twod = ""
                    for data_item in elt_geom:
                        coord = _try_text(xs.filepath, data_item).strip()[-1]
                        if coord in "XYZ":
                            twod += coord
                data_item = elt_geom[0]
                data_text = _try_text(xs.filepath, data_item)
                coord_shape = self._get_dims(data_item)
                coord_filepattern = data_text.strip().split(":/", 1)[0]
                coord_file_chunks = coord_filepattern.split("_")
                coord_file_chunks[-2] = "{icore:05d}"
                coord_filepattern = "_".join(coord_file_chunks)

            fields_info = {}
            for elt_fvar in xs.iter_load_successive_tag("Attribute"):
                name = elt_fvar.attrib["Name"]
                elt_data = elt_fvar[0]
                shape = self._get_dims(elt_data)
                ifile, isnap = _ifile_isnap(xs.filepath, elt_data)
                fields_info[name] = (ifile, shape)

            r_yin, r_yang = _count_subdomains(xs, i0_yin)

            data[isnap] = XmfEntry(
                isnap=isnap,
                time=time,
                mo_lambda=extra.get("mo_lambda"),
                mo_thick_sol=extra.get("mo_thick_sol"),
                yin_yang=yin_yang,
                twod=twod,
                coord_filepattern=coord_filepattern,
                coord_shape=coord_shape,
                range_yin=r_yin,
                range_yang=r_yang,
                fields=fields_info,
            )
        return data

    def __getitem__(self, isnap: int) -> XmfEntry:
        try:
            return self._data[isnap]
        except KeyError:
            raise ParsingError(self.path, f"no data for snapshot {isnap}")


def read_geom_h5(xdmf: FieldXmf, snapshot: int) -> dict[str, Any]:
    """Extract geometry information from hdf5 files.

    Args:
        xdmf: xdmf file parser.
        snapshot: snapshot number.

    Returns:
        geometry information.
    """
    header: dict[str, Any] = {}

    entry = xdmf[snapshot]
    header["ti_ad"] = entry.time
    header["mo_lambda"] = entry.mo_lambda
    header["mo_thick_sol"] = entry.mo_thick_sol
    header["ntb"] = 2 if entry.yin_yang else 1

    all_meshes: list[dict[str, NDArray[np.float64]]] = []
    for h5file in entry.coord_files_yin(xdmf.path.parent):
        all_meshes.append({})
        with h5py.File(h5file, "r") as h5f:
            for coord, mesh in h5f.items():
                # for some reason, the array is transposed!
                all_meshes[-1][coord] = mesh[()].reshape(entry.coord_shape).T
                all_meshes[-1][coord] = _make_3d(all_meshes[-1][coord], entry.twod)

    header["ncs"] = _ncores(all_meshes, entry.twod)
    header["nts"] = list(
        (all_meshes[0]["X"].shape[i] - 1) * header["ncs"][i] for i in range(3)
    )
    header["nts"] = np.array([max(1, val) for val in header["nts"]])
    meshes = _conglomerate_meshes(all_meshes, header)
    if np.any(meshes["Z"][:, :, 0] != 0):
        # spherical
        if entry.twod is not None:  # annulus geometry...
            header["x_mesh"] = np.copy(meshes["Y"])
            header["y_mesh"] = np.copy(meshes["Z"])
            header["z_mesh"] = np.copy(meshes["X"])
        else:  # YinYang, here only yin
            header["x_mesh"] = np.copy(meshes["X"])
            header["y_mesh"] = np.copy(meshes["Y"])
            header["z_mesh"] = np.copy(meshes["Z"])
        header["r_mesh"] = np.sqrt(
            header["x_mesh"] ** 2 + header["y_mesh"] ** 2 + header["z_mesh"] ** 2
        )
        header["t_mesh"] = np.arccos(header["z_mesh"] / header["r_mesh"])
        header["p_mesh"] = np.roll(
            np.arctan2(header["y_mesh"], -header["x_mesh"]) + np.pi, -1, 1
        )
        header["e1_coord"] = header["t_mesh"][:, 0, 0]
        header["e2_coord"] = header["p_mesh"][0, :, 0]
        header["e3_coord"] = header["r_mesh"][0, 0, :]
    else:
        header["e1_coord"] = meshes["X"][:, 0, 0]
        header["e2_coord"] = meshes["Y"][0, :, 0]
        header["e3_coord"] = meshes["Z"][0, 0, :]
    header["aspect"] = (
        header["e1_coord"][-1] - header["e2_coord"][0],
        header["e1_coord"][-1] - header["e2_coord"][0],
    )
    header["rcmb"] = header["e3_coord"][0]
    if header["rcmb"] == 0:
        header["rcmb"] = -1
    else:
        header["e3_coord"] = header["e3_coord"] - header["rcmb"]
    if entry.twod is None or "X" in entry.twod:
        header["e1_coord"] = header["e1_coord"][:-1]
    if entry.twod is None or "Y" in entry.twod:
        header["e2_coord"] = header["e2_coord"][:-1]
    header["e3_coord"] = header["e3_coord"][:-1]

    return header


def _to_spherical(
    flds: NDArray[np.float64], header: dict[str, Any]
) -> NDArray[np.float64]:
    """Convert vector field to spherical."""
    cth = np.cos(header["t_mesh"][:, :, :-1])
    sth = np.sin(header["t_mesh"][:, :, :-1])
    cph = np.cos(header["p_mesh"][:, :, :-1])
    sph = np.sin(header["p_mesh"][:, :, :-1])
    fout = np.copy(flds)
    fout[0] = cth * cph * flds[0] + cth * sph * flds[1] - sth * flds[2]
    fout[1] = sph * flds[0] - cph * flds[1]  # need to take the opposite here
    fout[2] = sth * cph * flds[0] + sth * sph * flds[1] + cth * flds[2]
    return fout


def _flds_shape(fieldname: str, header: dict[str, Any]) -> list[int]:
    """Compute shape of flds variable."""
    shp = list(header["nts"])
    shp.append(header["ntb"])
    if len(FIELD_FILES_H5.get(fieldname, [])) == 3:
        shp.insert(0, 3)
        # extra points
        header["xp"] = int(header["nts"][0] != 1)
        shp[1] += header["xp"]
        header["yp"] = int(header["nts"][1] != 1)
        shp[2] += header["yp"]
        header["zp"] = 1
        header["xyp"] = 1
    else:
        shp.insert(0, 1)
        header["xp"] = 0
        header["yp"] = 0
        header["zp"] = 0
        header["xyp"] = 0
    return shp


def _post_read_flds(
    flds: NDArray[np.float64], header: dict[str, Any]
) -> NDArray[np.float64]:
    """Process flds to handle sphericity."""
    if flds.shape[0] >= 3 and header["rcmb"] > 0:
        # spherical vector
        header["p_mesh"] = np.roll(
            np.arctan2(header["y_mesh"], header["x_mesh"]), -1, 1
        )
        for ibk in range(header["ntb"]):
            flds[..., ibk] = _to_spherical(flds[..., ibk], header)
        header["p_mesh"] = np.roll(
            np.arctan2(header["y_mesh"], -header["x_mesh"]) + np.pi, -1, 1
        )
    return flds


def read_field_h5(
    xdmf: FieldXmf,
    fieldname: str,
    snapshot: int,
    header: dict[str, Any] | None = None,
) -> tuple[dict[str, Any], NDArray[np.float64]] | None:
    """Extract field data from hdf5 files.

    Args:
        xdmf: xdmf file parser.
        fieldname: name of field to extract.
        snapshot: snapshot number.
        header: geometry information.

    Returns:
        geometry information and field data. None is returned if data is
            unavailable.
    """
    if header is None:
        header = read_geom_h5(xdmf, snapshot)

    npc = header["nts"] // header["ncs"]  # number of grid point per node
    flds = np.zeros(_flds_shape(fieldname, header))
    data_found = False

    for fsub in xdmf[snapshot].field_subdomains(xdmf.path.parent, fieldname):
        fld = _read_group_h5(fsub.file, fsub.dataset).reshape(fsub.shape)
        # for some reason, the field is transposed
        fld = fld.T
        shp = fld.shape

        if shp[-1] == 1 and header["nts"][0] == 1:  # YZ
            fld = fld.reshape((shp[0], 1, shp[1], shp[2]))
            if header["rcmb"] < 0:
                fld = fld[(2, 0, 1), ...]
        elif shp[-1] == 1:  # XZ
            fld = fld.reshape((shp[0], shp[1], 1, shp[2]))
            if header["rcmb"] < 0:
                fld = fld[(1, 2, 0), ...]
        elif fieldname in SFIELD_FILES_H5:
            fld = fld.reshape((1, npc[0], npc[1], 1))
        elif header["nts"][1] == 1:  # cart XZ
            fld = fld.reshape((1, shp[0], 1, shp[1]))
        ifs = [
            fsub.icore // np.prod(header["ncs"][:i]) % header["ncs"][i] * npc[i]
            for i in range(3)
        ]
        if fieldname in SFIELD_FILES_H5:
            ifs[2] = 0
            npc[2] = 1
        if header["zp"]:  # remove top row
            fld = fld[:, :, :, :-1]
        flds[
            :,
            ifs[0] : ifs[0] + npc[0] + header["xp"],
            ifs[1] : ifs[1] + npc[1] + header["yp"],
            ifs[2] : ifs[2] + npc[2],
            fsub.iblock,
        ] = fld
        data_found = True

    if flds.shape[0] == 3 and flds.shape[-1] == 2:  # YinYang vector
        # Yang grid is rotated compared to Yin grid
        flds[0, ..., 1] = -flds[0, ..., 1]
        vt = flds[1, ..., 1].copy()
        flds[1, ..., 1] = flds[2, ..., 1]
        flds[2, ..., 1] = vt
    flds = _post_read_flds(flds, header)

    if fieldname in SFIELD_FILES_H5:
        # remove z component
        flds = flds[..., 0, :]
    return (header, flds) if data_found else None


@dataclass(frozen=True)
class TracerSub:
    file: Path
    dataset: str
    icore: int
    iblock: int


@dataclass(frozen=True)
class XmfTracersEntry:
    isnap: int
    time: float | None
    mo_lambda: float | None
    mo_thick_sol: float | None
    yin_yang: bool
    range_yin: range
    range_yang: range
    fields: Mapping[str, int]

    def _fsub(self, path_root: Path, name: str, icore: int, yang: bool) -> TracerSub:
        ifile = self.fields[name]
        yy_tag = "2" if yang else "_"
        ic = icore + 1
        return TracerSub(
            file=path_root / f"Tracers_{name}_{ifile:05d}_{ic:05d}.h5",
            dataset=f"{name}{yy_tag}{ic:05d}_{self.isnap:05d}",
            icore=icore,
            iblock=int(yang),
        )

    def tra_subdomains(self, path_root: Path, name: str) -> Iterator[TracerSub]:
        if name not in self.fields:
            return
        for icore in self.range_yin:
            yield self._fsub(path_root, name, icore, False)
        for icore in self.range_yang:
            yield self._fsub(path_root, name, icore, True)


@dataclass(frozen=True)
class TracersXmf:
    path: Path

    @cached_property
    def _data(self) -> Mapping[int, XmfTracersEntry]:
        xs = XmlStream(filepath=self.path)
        data = {}
        for _ in xs.iter_tag("Time"):
            time = float(xs.current.attrib["Value"])
            xs.advance()
            extra: dict[str, float] = {}
            while xs.current.tag != "Grid":
                # mo_lambda, mo_thick_sol
                extra[xs.current.tag] = float(xs.current.attrib["Value"])
                xs.advance()

            mesh_name = xs.current.attrib["Name"]
            yin_yang = mesh_name.startswith("meshYin")
            i0_yin = int(mesh_name[-5:]) - 1

            fields_info = {}

            xs.skip_to_tag("Geometry")
            with xs.load() as elt_geom:
                for name, data_item in zip("zyx", elt_geom):
                    ifile, isnap = _ifile_isnap(xs.filepath, data_item)
                    fields_info[name] = ifile

            for elt_fvar in xs.iter_load_successive_tag("Attribute"):
                name = elt_fvar.attrib["Name"]
                ifile, _ = _ifile_isnap(xs.filepath, elt_fvar[0])
                fields_info[name] = ifile

            r_yin, r_yang = _count_subdomains(xs, i0_yin)

            data[isnap] = XmfTracersEntry(
                isnap=isnap,
                time=time,
                mo_lambda=extra.get("mo_lambda"),
                mo_thick_sol=extra.get("mo_thick_sol"),
                yin_yang=yin_yang,
                range_yin=r_yin,
                range_yang=r_yang,
                fields=fields_info,
            )
        return data

    def __getitem__(self, isnap: int) -> XmfTracersEntry:
        try:
            return self._data[isnap]
        except KeyError:
            raise ParsingError(self.path, f"no data for snapshot {isnap}")


def read_tracers_h5(
    xdmf: TracersXmf, infoname: str, snapshot: int
) -> list[NDArray[np.float64]]:
    """Extract tracers data from hdf5 files.

    Args:
        xdmf: xdmf file parser.
        infoname: name of information to extract.
        snapshot: snapshot number.

    Returns:
        Tracers data organized by attribute and block.
    """
    tra: list[list[NDArray[np.float64]]] = [[], []]  # [block][core]
    for tsub in xdmf[snapshot].tra_subdomains(xdmf.path.parent, infoname):
        tra[tsub.iblock].append(_read_group_h5(tsub.file, tsub.dataset))

    tra_concat: list[NDArray[np.float64]] = []
    for trab in tra:
        if trab:
            tra_concat.append(np.concatenate(trab))
    return tra_concat


def read_time_h5(h5folder: Path) -> Iterator[tuple[int, int]]:
    """Iterate through (isnap, istep) recorded in h5folder/'time_botT.h5'.

    Args:
        h5folder: directory of HDF5 output files.

    Yields:
        tuple (isnap, istep).
    """
    with h5py.File(h5folder / "time_botT.h5", "r") as h5f:
        for name, dset in h5f.items():
            isnap = int(name[-5:])
            if len(dset) == 3:
                istep = int(dset[2])
            else:
                istep = int(dset[0])
            yield isnap, istep



"""Implementation of Step objects.

Note:
    This module and the classes it defines are internals of StagPy, they
    should not be used in an external script.  Instead, use the
    [`StagyyData`][stagpy.stagyydata.StagyyData] class.
"""

from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import cached_property

import numpy as np

from . import error, phyvars, stagyyparsers
from .datatypes import Field, Rprof, Varr
from .dimensions import Scales

if typing.TYPE_CHECKING:
    from collections.abc import Mapping
    from typing import Any, Callable, NoReturn

    from numpy.typing import NDArray
    from pandas import DataFrame, Series

    from ._caching import FieldCache
    from .datatypes import Varf
    from .stagyydata import StagyyData


@dataclass(frozen=True)
class Geometry:
    """Geometry information.

    It is deduced from the information in the header of binary field files
    output by StagYY.
    """

    step: Step

    @cached_property
    def _maybe_header(self) -> dict[str, Any] | None:
        if self.step.isnap is None:
            return None
        sdat = self.step.sdat
        binfiles = sdat._binfiles_set(self.step.isnap)
        header = None
        if binfiles:
            header = stagyyparsers.field_header(binfiles.pop())
        elif sdat.hdf5:
            header = stagyyparsers.read_geom_h5(sdat._dataxmf, self.step.isnap)
        return header if header else None

    @cached_property
    def _header(self) -> Mapping[str, Any]:
        if self._maybe_header is None:
            raise error.NoGeomError(self.step)
        return self._maybe_header

    def _scale_radius_mo(self, radius: NDArray[np.float64]) -> NDArray[np.float64]:
        """Rescale radius for evolving MO runs."""
        if self.step.sdat.par.get("magma_oceans_in", "evolving_magma_oceans", False):
            return self._header["mo_thick_sol"] * (radius + self._header["mo_lambda"])
        return radius

    @property
    def aspect_ratio(self) -> tuple[float, float]:
        return self.step.sdat.par.nml["geometry"]["aspect_ratio"]

    @cached_property
    def _ntot(self) -> tuple[int, int, int, int]:
        return (*self._header["nts"], self._header["ntb"])

    @cached_property
    def nttot(self) -> int:
        """Number of grid point along the x/theta direction."""
        return self._ntot[0]

    @cached_property
    def nptot(self) -> int:
        """Number of grid point along the y/phi direction."""
        return self._ntot[1]

    @cached_property
    def nrtot(self) -> int:
        """Number of grid point along the z/r direction."""
        return self._ntot[2]

    @cached_property
    def nbtot(self) -> int:
        """Number of blocks."""
        return self._ntot[3]

    @property
    def nxtot(self) -> int:
        """Same as nttot."""
        return self.nttot

    @property
    def nytot(self) -> int:
        """Same as nptot."""
        return self.nptot

    @property
    def nztot(self) -> int:
        """Same as nrtot."""
        return self.nrtot

    @cached_property
    def r_walls(self) -> NDArray[np.float64]:
        """Position of FV walls along the z/r direction."""
        rgeom = self._header.get("rgeom")
        if rgeom is not None:
            walls = rgeom[:, 0] + self.rcmb
        else:
            walls = self._header["e3_coord"] + self.rcmb
            walls = np.append(walls, self.step.rprofs.bounds[1])
        return self._scale_radius_mo(walls)

    @cached_property
    def r_centers(self) -> NDArray[np.float64]:
        """Position of FV centers along the z/r direction."""
        rgeom = self._header.get("rgeom")
        if rgeom is not None:
            walls = rgeom[:-1, 1] + self.rcmb
        else:
            walls = self.step.rprofs.centers
        return self._scale_radius_mo(walls)

    @cached_property
    def t_walls(self) -> NDArray[np.float64]:
        """Position of FV walls along x/theta."""
        if self.threed or self.twod_xz:
            if self.yinyang:
                tmin, tmax = -np.pi / 4, np.pi / 4
            elif self.curvilinear:
                # should take theta_position/theta_center into account
                tmin = 0
                tmax = min(np.pi, self.aspect_ratio[0])
            else:
                tmin = 0
                tmax = self.aspect_ratio[0]
            return np.linspace(tmin, tmax, self.nttot + 1, dtype=np.float64)
        # twoD YZ
        center = np.pi / 2 if self.curvilinear else 0
        d_t = (self.p_walls[1] - self.p_walls[0]) / 2
        return np.array([center - d_t, center + d_t])

    @cached_property
    def t_centers(self) -> NDArray[np.float64]:
        """Position of FV centers along x/theta."""
        return (self.t_walls[:-1] + self.t_walls[1:]) / 2  # type: ignore

    @cached_property
    def p_walls(self) -> NDArray[np.float64]:
        """Position of FV walls along y/phi."""
        if self.threed or self.twod_yz:
            if self.yinyang:
                pmin, pmax = -3 * np.pi / 4, 3 * np.pi / 4
            elif self.curvilinear:
                pmin = 0
                pmax = min(2 * np.pi, self.aspect_ratio[1])
            else:
                pmin = 0
                pmax = self.aspect_ratio[1]
            return np.linspace(pmin, pmax, self.nptot + 1, dtype=np.float64)
        # twoD YZ
        d_p = (self.t_walls[1] - self.t_walls[0]) / 2
        return np.array([-d_p, d_p])

    @cached_property
    def p_centers(self) -> NDArray[np.float64]:
        """Position of FV centers along y/phi."""
        return (self.p_walls[:-1] + self.p_walls[1:]) / 2  # type: ignore

    @property
    def z_walls(self) -> NDArray[np.float64]:
        """Same as r_walls."""
        return self.r_walls

    @property
    def z_centers(self) -> NDArray[np.float64]:
        """Same as r_centers."""
        return self.r_centers

    @property
    def x_walls(self) -> NDArray[np.float64]:
        """Same as t_walls."""
        return self.t_walls

    @property
    def x_centers(self) -> NDArray[np.float64]:
        """Same as t_centers."""
        return self.t_centers

    @property
    def y_walls(self) -> NDArray[np.float64]:
        """Same as p_walls."""
        return self.p_walls

    @property
    def y_centers(self) -> NDArray[np.float64]:
        """Same as p_centers."""
        return self.p_centers

    @cached_property
    def rcmb(self) -> float:
        """Radius of CMB, 0 in cartesian geometry."""
        return max(self._header["rcmb"], 0)

    @cached_property
    def _shape(self) -> str:
        return self.step.sdat.par.nml["geometry"]["shape"].lower()

    @cached_property
    def curvilinear(self) -> bool:
        """Whether the grid is in curvilinear geometry."""
        return self._header["rcmb"] >= 0

    @property
    def cartesian(self) -> bool:
        """Whether the grid is in cartesian geometry."""
        return not self.curvilinear

    @cached_property
    def cylindrical(self) -> bool:
        """Whether the grid is in cylindrical geometry (2D spherical)."""
        aspect = self._header["aspect"]
        return (
            self.curvilinear
            and self.twod_xz
            and (self._shape == "cylindrical" or aspect[0] >= np.pi)
        )

    @cached_property
    def spherical(self) -> bool:
        """Whether the grid is in spherical geometry."""
        return self.curvilinear and not self.cylindrical

    @cached_property
    def axisymmetric(self) -> bool:
        """Whether the grid is in cartesian axisymmetric geometry."""
        return self.cartesian and self.twod_xz and self._shape == "axisymmetric"

    @property
    def yinyang(self) -> bool:
        """Whether the grid is in Yin-yang geometry (3D spherical)."""
        return self.spherical and self.nbtot == 2

    @property
    def twod_xz(self) -> bool:
        """Whether the grid is in the XZ plane only."""
        return self.nytot == 1

    @property
    def twod_yz(self) -> bool:
        """Whether the grid is in the YZ plane only."""
        return self.nxtot == 1

    @property
    def twod(self) -> bool:
        """Whether the grid is 2 dimensional."""
        return self.twod_xz or self.twod_yz

    @property
    def threed(self) -> bool:
        """Whether the grid is 3 dimensional."""
        return not self.twod

    def at_z(self, zval: float) -> int:
        """Return iz closest to given zval position.

        In spherical geometry, the bottom boundary is considered to be at z=0.
        Use `at_r` to find a cell at a given radial position.
        """
        if self.curvilinear:
            zval += self.rcmb
        return int(np.argmin(np.abs(self.z_centers - zval)))

    def at_r(self, rval: float) -> int:
        """Return ir closest to given rval position.

        If called in cartesian geometry, this is equivalent to `at_z`.
        """
        return int(np.argmin(np.abs(self.r_centers - rval)))


@dataclass(frozen=True)
class Fields:
    """Fields data structure.

    The `Step.fields` attribute is an instance of this class.
    """

    step: Step
    variables: Mapping[str, Varf]
    extravars: Mapping[str, Callable[[Step], Field]]
    files: Mapping[str, list[str]]
    filesh5: Mapping[str, list[str]]

    @cached_property
    def _all_vars(self) -> set[str]:
        return set(self.variables.keys()).union(self.extravars.keys())

    @cached_property
    def _cache(self) -> FieldCache:
        return self.step.sdat._field_cache

    def __getitem__(self, name: str) -> Field:
        if name not in self._all_vars:
            raise error.UnknownFieldVarError(name)

        maybe_fld = self._cache.get(self.step.istep, name)
        if maybe_fld is not None:
            return maybe_fld

        if name in self.extravars:
            fld = self.extravars[name](self.step)
            self._cache.insert(self.step.istep, name, fld)
            return fld

        # requested field is one of self._vars
        fld_names, parsed_data = self._get_raw_data(name)
        if parsed_data is None:
            raise error.MissingDataError(
                f"Missing field {name} in step {self.step.istep}"
            )
        _, fields = parsed_data
        for fld_name, fld_vals in zip(fld_names, fields):
            fld = Field(fld_vals, self.variables[fld_name])
            self._cache.insert(self.step.istep, fld_name, fld)
        return self[name]

    def __contains__(self, item: Any) -> bool:
        try:
            _ = self[item]
        except error.MissingDataError:
            return False
        return True

    def _get_raw_data(self, name: str) -> tuple[list[str], Any]:
        """Find file holding data and return its content."""
        # try legacy first, then hdf5
        filestem = ""
        for filestem, list_fvar in self.files.items():
            if name in list_fvar:
                break
        parsed_data = None
        if self.step.isnap is None:
            return list_fvar, None
        fieldfile = self.step.sdat.filename(
            filestem, self.step.isnap, force_legacy=True
        )
        if not fieldfile.is_file():
            fieldfile = self.step.sdat.filename(filestem, self.step.isnap)
        if fieldfile.is_file():
            parsed_data = stagyyparsers.fields(fieldfile)
        elif self.step.sdat.hdf5 and self.filesh5:
            # files in which the requested data can be found
            files = [
                (stem, fvars) for stem, fvars in self.filesh5.items() if name in fvars
            ]
            for filestem, list_fvar in files:
                sdat = self.step.sdat
                if filestem in phyvars.SFIELD_FILES_H5:
                    xmff = sdat._botxmf if name.endswith("bot") else sdat._topxmf
                    header = self.step.geom._maybe_header
                    assert header is not None
                else:
                    xmff = sdat._dataxmf
                    header = None
                parsed_data = stagyyparsers.read_field_h5(
                    xmff, filestem, self.step.isnap, header
                )
                if parsed_data is not None:
                    break
        return list_fvar, parsed_data


@dataclass(frozen=True)
class Tracers:
    """Tracers data structure.

    The `Step.tracers` attribute is an instance of this class.

    `Tracers` implements the getitem mechanism. Items are tracervar names such
    as `"Type"` or `"Mass"`.  The position of tracers are the `"x"`, `"y"` and
    `"z"` items.
    """

    step: Step

    @cached_property
    def _data(self) -> dict[str, list[NDArray[np.floating]] | None]:
        return {}

    def __getitem__(self, name: str) -> list[NDArray[np.floating]] | None:
        if name in self._data:
            return self._data[name]
        if self.step.isnap is None:
            return None
        data = stagyyparsers.tracers(
            self.step.sdat.filename("tra", timestep=self.step.isnap, force_legacy=True)
        )
        if data is None and self.step.sdat.hdf5:
            self._data[name] = stagyyparsers.read_tracers_h5(  # type: ignore
                self.step.sdat._traxmf,
                name,
                self.step.isnap,
            )
        elif data is not None:
            self._data.update(data)
        if name not in self._data:
            self._data[name] = None
        return self._data[name]

    def __iter__(self) -> NoReturn:
        raise TypeError("tracers collection is not iterable")


class Rprofs(ABC):
    """Radial profiles.

    `Rprofs` implements the getitem mechanism.  Keys are profile names
    defined in `stagpy.phyvars.RPROF[_EXTRA]`.  Items are
    [`Rprof`][stagpy.datatypes.Rprof] instances.
    """

    @abstractmethod
    def __getitem__(self, name: str) -> Rprof: ...

    @property
    @abstractmethod
    def stepstr(self) -> str:
        """String representation of steps indices."""

    @property
    @abstractmethod
    def centers(self) -> NDArray[np.float64]:
        """Radial position of cell centers."""

    @property
    @abstractmethod
    def walls(self) -> NDArray[np.float64]:
        """Radial position of cell walls."""

    @property
    @abstractmethod
    def bounds(self) -> tuple[float, float]:
        """Radial or vertical position of boundaries.

        Radial/vertical positions of boundaries of the domain.
        """


@dataclass(frozen=True)
class RprofsInstant(Rprofs):
    """Radial profiles at a given step.

    The `Step.rprofs` attribute is an instance of this class.
    """

    step: Step

    @cached_property
    def _cached_extra(self) -> dict[str, Rprof]:
        return {}

    @cached_property
    def _data(self) -> DataFrame | None:
        step = self.step
        return step.sdat._rprof_and_times[0].get(step.istep)

    @property
    def _rprofs(self) -> DataFrame:
        if self._data is None:
            step = self.step
            raise error.MissingDataError(
                f"No rprof data in step {step.istep} of {step.sdat}"
            )
        return self._data

    def __getitem__(self, name: str) -> Rprof:
        step = self.step
        rprof: NDArray[np.float64]
        if name in self._rprofs.columns:
            rprof = self._rprofs[name].to_numpy(dtype=np.float64)
            rad = self.centers
            if name in phyvars.RPROF:
                meta = phyvars.RPROF[name]
            else:
                meta = Varr(name, "", "1")
        elif name in self._cached_extra:
            rpf = self._cached_extra[name]
            rprof = rpf.values
            rad = rpf.rad
            meta = rpf.meta
        elif name in phyvars.RPROF_EXTRA:
            self._cached_extra[name] = phyvars.RPROF_EXTRA[name](step)
            rpf = self._cached_extra[name]
            rprof = rpf.values
            rad = rpf.rad
            meta = rpf.meta
        else:
            raise error.UnknownRprofVarError(name)

        return Rprof(rprof, rad, meta)

    @property
    def stepstr(self) -> str:
        """String representation of the parent :class:`Step`."""
        return str(self.step.istep)

    @cached_property
    def centers(self) -> NDArray[np.float64]:
        """Radial position of cell centers."""
        return self._rprofs["r"].to_numpy() + self.bounds[0]  # type: ignore

    @cached_property
    def walls(self) -> NDArray[np.float64]:
        """Radial position of cell walls."""
        rbot, rtop = self.bounds
        try:
            walls = self.step.geom.r_walls
        except error.StagpyError:
            # assume walls are mid-way between T-nodes
            # could be T-nodes at center between walls
            centers = self.centers
            walls: NDArray[np.float64] = (centers[:-1] + centers[1:]) / 2  # type: ignore
            walls = np.insert(walls, 0, rbot)
            walls = np.append(walls, rtop)
        return walls

    @cached_property
    def bounds(self) -> tuple[float, float]:
        """Radial or vertical position of boundaries.

        Radial/vertical positions of boundaries of the domain.
        """
        step = self.step
        try:
            rcmb = step.geom.rcmb
        except error.NoGeomError:
            rcmb = step.sdat.par.get("geometry", "r_cmb", 3480e3)
            if step.sdat.par.nml["geometry"]["shape"].lower() == "cartesian":
                rcmb = 0
        rbot = max(rcmb, 0)
        thickness = (
            Scales(step.sdat).length
            if step.sdat.par.get("switches", "dimensional_units", True)
            else 1
        )
        return rbot, rbot + thickness


@dataclass(frozen=True)
class Step:
    """Time step data structure.

    Elements of [`Steps`][stagpy.stagyydata.Steps] and
    [`Snaps`][stagpy.stagyydata.Snaps] instances are all `Step`
    instances. Note that `Step` objects are not duplicated.

    Examples:
        Here are a few examples illustrating some properties of `Step`
        instances.

        ```py
        sdat = StagyyData(Path("path/to/run"))
        istep_last_snap = sdat.snaps[-1].istep
        assert(sdat.steps[istep_last_snap] is sdat.snaps[-1])
        n = 0  # or any valid time step / snapshot index
        assert(sdat.steps[n].sdat is sdat)
        assert(sdat.steps[n].istep == n)
        assert(sdat.snaps[n].isnap == n)
        assert(sdat.snaps[n] is sdat.snaps[n].fields.step)
        ```

    Attributes:
        istep: the index of the time step that the instance represents.
        sdat: the owner of the `Step` instance.
    """

    istep: int
    sdat: StagyyData

    @cached_property
    def fields(self) -> Fields:
        """Fields available at this time step."""
        return Fields(
            self,
            phyvars.FIELD,
            phyvars.FIELD_EXTRA,
            phyvars.FIELD_FILES,
            phyvars.FIELD_FILES_H5,
        )

    @cached_property
    def sfields(self) -> Fields:
        """Surface fields available at this time step."""
        return Fields(
            self,
            phyvars.SFIELD,
            {},
            phyvars.SFIELD_FILES,
            phyvars.SFIELD_FILES_H5,
        )

    @cached_property
    def tracers(self) -> Tracers:
        """Tracer information available at this time step."""
        return Tracers(self)

    @cached_property
    def rprofs(self) -> RprofsInstant:
        """Radial profiles available at this time step."""
        return RprofsInstant(self)

    def __repr__(self) -> str:
        if self.isnap is not None:
            return f"{self.sdat!r}.snaps[{self.isnap}]"
        else:
            return f"{self.sdat!r}.steps[{self.istep}]"

    @cached_property
    def geom(self) -> Geometry:
        """Geometry information.

        [`Geometry`][stagpy.step.Geometry] instance holding geometry information.
        It is issued from binary files holding field information.
        """
        return Geometry(self)

    @property
    def timeinfo(self) -> Series[np.float64]:
        """Time series data of the time step."""
        try:
            info = self.sdat.tseries.at_step(self.istep)
        except KeyError:
            raise error.MissingDataError(f"No time series for {self!r}")
        return info

    @property
    def time(self) -> float:
        """Time of this time step."""
        steptime = None
        try:
            steptime = self.timeinfo["time"]
        except error.MissingDataError:
            if self.isnap is not None:
                steptime = self.geom._header.get("ti_ad")
        if steptime is None:
            raise error.NoTimeError(self)
        return steptime

    @property
    def isnap(self) -> int | None:
        """Snapshot index corresponding to time step.

        It is None if no snapshot exists for the time step.
        """
        return self.sdat._step_snap.isnap(istep=self.istep)



"""StagPy is a tool to postprocess StagYY output files.

StagPy is both a CLI tool and a powerful Python library. See the
documentation at https://stagpython.github.io/StagPy/

When using the CLI interface, warnings are ignored and only a short form of
encountered StagpyErrors is printed. Set the environment variable STAGPY_DEBUG
to issue warnings normally and raise StagpyError.
"""

from importlib.metadata import version

__version__ = version("stagpy")
